import typing as T

import numpy as np
import pandas as pd
import xarray as xr


def _return_sorted_array(values: T.Union[float, int, np.ndarray, T.List]) -> np.ndarray:
    """Sort values or ensure value is an array"""
    if isinstance(values, (T.List, np.ndarray)):
        return np.sort(values)
    else:
        return np.array([values])


def digitize(
    data: T.Union[xr.Dataset, xr.DataArray, pd.DataFrame, pd.Series, np.ndarray, T.List] = None,  # noqa: C901
    bins: T.Union[float, int, np.ndarray, T.List] = [5, 10],
    closed_right: bool = False,
    columns: T.Union[str, np.ndarray, T.List] = None,
    keep_attrs: bool = True,
) -> T.Union[xr.Dataset, xr.DataArray, pd.DataFrame, pd.Series, np.ndarray, T.List]:
    """
    Return the indices of the bins to which each value in the input data belongs. This function is essentially a
    wrapper around numpy.digitize, and provides consistent usage for a variety of data structures.
    Automatically sorts the bins from lowest to highest. Accepts single values for bins or an array/list.

    ==============  ============================
    `closed_right`  returned index `i` satisfies
    ==============  ============================
    ``False``       ``bins[i-1] <= x < bins[i]``
    ``True``        ``bins[i-1] < x <= bins[i]``
    ==============  ============================

    If values are beyond the bounds of `bins`, 0 or ``len(bins)`` is returned as appropriate.

    A potential use case is for labelling values for scoring using a confusion matrix.

    Args:
        data: Input data
        bins: Array/list of values to use for thresholds.
        closed_right: Indicating whether the intervals include the right or the left bin edge.
        By default, closed_right=False, meaning that the interval does not include the right edge.
        columns: List of column names for which to apply the thresholding. Only used for pandas.Dataframe type.
        keep_attrs: Keep attributes if data is of type xr.DataArray. Ignored for other types of data.

    Returns:
        digitized_data: The data (same type as input) with values replaced by integers corresponding
        to indices of the given bins.

    >>> array = np.arange(1, 8, 0.5)
    >>> array
    array([1. , 1.5, 2. , 2.5, 3. , 3.5, 4. , 4.5, 5. , 5.5, 6. , 6.5, 7. ,
           7.5])
    >>> digitize(array, bins=[3, 6], closed_right=False)
    array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2])
    >>> digitize(array, bins=[6, 3], closed_right=True)
    array([0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 2, 2, 2])
    >>> digitize(array, bins=3)
    array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1])
    """
    bins = _return_sorted_array(bins)

    if isinstance(data, (np.ndarray, T.List)):
        digitized_data = np.digitize(data, bins=bins, right=closed_right)

    elif isinstance(data, pd.Series):
        result = np.digitize(data, bins=bins, right=True)
        digitized_data = pd.Series(result, index=data.index)

    elif isinstance(data, pd.DataFrame):
        if columns is None:
            dataframe = data
        else:
            # If column names(s) are given, ensure names are a list so a dataframe is always created
            if isinstance(columns, str):
                columns = [columns]
            dataframe = data[columns]

        # Apply digitize to dataframe
        result = dataframe.apply(np.digitize, axis=1, result_type="expand", bins=bins, right=closed_right)
        result.index = data.index
        if columns is None:
            result.columns = data.columns
        else:
            result.columns = columns
        digitized_data = result

    elif isinstance(data, (xr.DataArray, xr.Dataset)):
        digitized_data = xr.apply_ufunc(
            np.digitize,
            data,
            kwargs={"bins": bins, "right": closed_right},
            keep_attrs=keep_attrs,
            dask="allowed",
            output_dtypes=["i8"],
        )
    else:
        allowed_types = ", ".join(
            "{}".format(obj) for obj in [xr.Dataset, xr.DataArray, pd.DataFrame, pd.Series, np.ndarray, list]
        )
        raise TypeError("'data' is of type {}, but must be one of type {}.".format(type(data), allowed_types))
    return digitized_data



import typing as T

import bottleneck as bn
import numpy as np
import pandas as pd
import xarray as xr


def calculate_quantiles(
    data: T.Union[xr.Dataset, xr.DataArray, pd.DataFrame, pd.Series] = None,
    q: T.Iterable = [0.25, 0.5, 0.75],
    dim: T.Union[str, int] = None,
    numeric_only: bool = False,
    interpolation: str = "linear",
    keep_attrs: bool = False,
    skipna: bool = True,
    add_rank_coord: bool = False,
    starting_rank: int = 1,
) -> T.Union[xr.Dataset, xr.DataArray, pd.DataFrame, pd.Series]:
    """
    Calculate quantiles for data over a given dimension. Essentially a wrapper around `xarray` and `pandas`
    functions, with the option for adding a `rank` coordinate:
    http://xarray.pydata.org/en/stable/generated/xarray.DataArray.quantile.html
    https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.quantile.html

    Args:
        data: Input data
        q: Levels at which to calculate the quantiles - must be between 0 and 1 inclusive.
        dim: For `xarray`, the name of dimension over which to calculate quantiles.
        For `pandas`, the axis number (one of `0`, `1`, `'index'`, `'columns'`), with default==0.
        numeric_only: Default==True. If False, the quantile of datetime and timedelta data will be computed as well.
        interpolation: Type of interpolation to use when quantile lies between two data points,
        one of 'linear', 'lower', 'higher', 'midpoint', 'nearest'.
        keep_attrs: Only for `xarray` objects. If True, the dataset’s attributes (attrs) will be copied from the
        original object to the new one. If False (default), the new object will be returned without attributes.
        skipna: Whether to skip missing values when aggregating. Default=True
        add_rank_coord: If True, add a coordinate (`xarray`) or column (`pandas`) with the rank
        value of the quantile (`range(len(q))`)
        starting_rank: Starting value of the rank coordinate/column

    Returns:
        Data with the values of the specified quantiles
    """
    allowed_types = (xr.Dataset, xr.DataArray, pd.DataFrame, pd.Series)
    if not isinstance(data, allowed_types):
        raise TypeError("`data` must be of type: {:s}".format(", ".join(str(obj) for obj in allowed_types)))

    if add_rank_coord:
        rank = np.arange(starting_rank, starting_rank + len(q), dtype=int)

    if isinstance(data, (xr.Dataset, xr.DataArray)):
        if isinstance(data, xr.DataArray):
            quantiles = data.quantile(q=q, dim=dim, interpolation=interpolation, keep_attrs=keep_attrs, skipna=skipna)
        else:
            quantiles = data.quantile(
                q=q,
                dim=dim,
                interpolation=interpolation,
                keep_attrs=keep_attrs,
                numeric_only=numeric_only,
                skipna=skipna,
            )

        if add_rank_coord:
            quantiles = quantiles.assign_coords(rank=("quantile", rank))

    elif isinstance(data, pd.DataFrame):
        if dim is None or isinstance(dim, str):
            dim = 0
        # Ensure `q` is a list so we always have an index
        if isinstance(q, (float, int)):
            q = [q]
        quantiles = data.quantile(q=q, axis=dim, numeric_only=numeric_only)
        if add_rank_coord:
            quantiles["rank"] = rank

    elif isinstance(data, pd.Series):
        if dim is None or isinstance(dim, str):
            dim = 0
        if add_rank_coord:
            # Ensure `q` is a list so we always have an index
            if isinstance(q, (float, int)):
                q = [q]
            quantiles = data.quantile(q=q, interpolation=interpolation)
            # Transform the Series to a DataFrame so we can have a rank column
            if quantiles.name is None:
                quantiles.name = "value"
            quantiles = quantiles.to_frame()
            quantiles["rank"] = rank
        else:
            quantiles = data.quantile(q=q, interpolation=interpolation)

    return quantiles


def rank(
    data: T.Union[xr.Dataset, xr.DataArray] = None, dim: str = "time", percent: bool = False, starting_rank: int = 1
) -> T.Union[xr.Dataset, xr.DataArray]:
    """
    Ranks data over the selected dimension. Equal values are assigned a rank
    that is the average of the ranks that would have been otherwise assigned
    to all of the values within that set. Replicates `xr.DataArray.rank` but
    with support for dask arrays - xarray documentation:
    http://xarray.pydata.org/en/stable/generated/xarray.DataArray.rank.html

    Can optionally set the starting rank to begin at 0 instead of the default of 1
    (the default from the `bottleneck` ranking function). NaNs in the input array
    are returned as NaNs.

    Args:
        data: Input data
        dim: Dimension over which to compute the rank
        percent: Optional. If `True`, compute percentage ranks, otherwise compute integer ranks.
        starting_rank: Starting number for the ranks if computing integer ranks.

    Returns:
        Ranked data with the same coordinates.
    """

    def _nan_rank(data):
        ranked = bn.nanrankdata(data, axis=-1)
        if percent:
            # Divide by the number of non-NaN values
            count = np.count_nonzero(~np.isnan(data), axis=-1, keepdims=True)
            ranked = ranked / count
        return ranked

    # Check if data is stored as Dask array, and if so, re-chunk over the ranking dimension
    if data.chunks is not None:
        if len(data.chunks) > 0:
            data = data.chunk({dim: -1})

    ranked_data = xr.apply_ufunc(
        _nan_rank,
        data,
        input_core_dims=[[dim]],
        output_core_dims=[[dim]],
        dask="parallelized",
        output_dtypes=[np.float32],
    )

    if percent:
        return ranked_data
    else:
        if starting_rank != 1:
            # By default rank starts at 1, so if choosing another
            # starting number, we need to add it and subtract 1
            ranked_data = ranked_data - 1 + starting_rank
        # As rank values are integers, cast to `int` dtype
        return ranked_data.astype(int)



from pathlib import Path
from unittest import TestCase

import geopandas as gpd
import numpy as np
import pandas as pd
import shapely
import xarray as xr

import terrapyn as tp

PACKAGE_ROOT_DIR = Path(__file__).resolve().parent.parent
TEST_DATA_PATH = PACKAGE_ROOT_DIR / "tests" / "data"


class TestGetDataAtCoords(TestCase):

    np.random.seed(42)
    n_lat = 4
    n_lon = 4
    n_time = 3
    data = 5 + np.random.randn(n_time, n_lat, n_lon)
    da = xr.DataArray(
        data,
        dims=["time", "lat", "lon"],
        coords={
            "time": pd.date_range("2014-09-06", periods=n_time),
            "lat": 3 + np.arange(n_lat),
            "lon": 13 + np.arange(n_lon),
        },
        name="var",
    )
    ds = da.to_dataset()

    lats = [3.32, 4.67]
    lons = [15.02, 15.73]
    ids = ["a", "b"]

    def test_nearest_coordinate_values(self):
        df = tp.space.get_data_at_coords(self.ds, lats=self.lats, lons=self.lons)
        nearest_values_point_0 = df.loc[(slice(None), 0), "var"].values
        np.testing.assert_almost_equal(nearest_values_point_0, np.array([5.64768854, 4.09197592, 5.82254491]))
        nearest_values_point_1 = df.loc[(slice(None), 1), "var"].values
        np.testing.assert_almost_equal(nearest_values_point_1, np.array([4.53427025, 5.37569802, 4.6988963]))

    def test_missing_latitudes(self):
        self.assertRaises(
            ValueError,
            tp.space.get_data_at_coords,
            ds=self.ds,
            lats=None,
            lons=self.lons,
        )

    def test_linear_values(self):
        df = tp.space.get_data_at_coords(self.ds, lats=self.lats, lons=self.lons, method="linear")
        nearest_values_point_0 = df.loc[(slice(None), 0), "var"].values
        np.testing.assert_almost_equal(nearest_values_point_0, np.array([5.95248557, 4.38774388, 5.11628122]))
        nearest_values_point_1 = df.loc[(slice(None), 1), "var"].values
        np.testing.assert_almost_equal(nearest_values_point_1, np.array([5.01396221, 4.63833409, 4.7608919]))

    def test_id_names(self):
        df = tp.space.get_data_at_coords(self.ds, lats=self.lats, lons=self.lons, point_names=self.ids, method="linear")
        self.assertEqual(
            list(df.index.get_level_values("id")),
            ["a", "b", "a", "b", "a", "b"],
        )

    def test_point_names_dim(self):
        df = tp.space.get_data_at_coords(
            self.ds,
            lats=self.lats,
            lons=self.lons,
            point_names=self.ids,
            method="linear",
            point_names_dim="test",
        )
        self.assertEqual(list(df.index.names), ["time", "test"])

    def test_dataset_without_time(self):
        data = self.ds.isel(time=0).drop_vars("time")
        df = tp.space.get_data_at_coords(data, lats=self.lats, lons=self.lons, method="nearest")
        self.assertEqual(df.index.name, "id")
        np.testing.assert_almost_equal(df["var"].values, np.array([5.64768854, 4.53427025]))


class TestBBox(TestCase):

    min_lon = 10
    max_lon = 20
    min_lat = 30
    max_lat = 40
    shapely_rectangle = shapely.geometry.box(minx=min_lon, miny=min_lat, maxx=max_lon, maxy=max_lat)
    shapely_polygon = shapely.geometry.polygon.Polygon([(0, 0), (1, 1), (2, 2), (2, 1), (0, 0)])

    def test_bounds(self):
        assert tp.space.BBox(geometry=self.shapely_rectangle).bounds == (10.0, 30.0, 20.0, 40.0)
        assert tp.space.BBox(geometry=self.shapely_polygon).bounds == (0.0, 0.0, 2.0, 2.0)

    def test_WNES(self):
        assert tp.space.BBox(
            min_lat=self.min_lat, max_lat=self.max_lat, min_lon=self.min_lon, max_lon=self.max_lon
        ).WNES == (self.min_lon, self.max_lat, self.max_lon, self.min_lat)

    def test_NWSE(self):
        assert tp.space.BBox(
            min_lat=self.min_lat, max_lat=self.max_lat, min_lon=self.min_lon, max_lon=self.max_lon
        ).NWSE == (self.max_lat, self.min_lon, self.min_lat, self.max_lon)

    def test_centroid(self):
        bbox = tp.space.BBox(min_lat=self.min_lat, max_lat=self.max_lat, min_lon=self.min_lon, max_lon=self.max_lon)
        assert isinstance(bbox.centroid, shapely.geometry.point.Point)
        self.assertEqual(bbox.centroid.coords[0], (15.0, 35.0))
        self.assertEqual(bbox.centroid_coords, (35.0, 15.0))

    def test_buffer(self):
        assert tp.space.BBox(geometry=self.shapely_polygon).buffer(1).bounds == (
            -0.99997179491729,
            -0.9997461638018406,
            3.0,
            3.0,
        )


class TestCropToBBox(TestCase):
    def test_pandas_dataframe(self):
        df = pd.DataFrame(
            {
                "time": pd.date_range("2019-03-15", freq="1D", periods=3),
                "id": [123, 456, 789],
                "lat": [1, 3, 5],
                "lon": [10, 20, 30],
            }
        ).set_index(["time", "id"])
        bbox = tp.space.BBox(max_lat=4, min_lat=0, max_lon=25, min_lon=0)
        results = tp.space.crop_to_bbox(df, bbox=bbox).values
        expected = np.array([[1, 10], [3, 20]])
        np.testing.assert_equal(results, expected)

    def test_geopandas_geodataframe(self):
        df = pd.DataFrame(
            {
                "time": pd.date_range("2019-03-15", freq="1D", periods=3),
                "id": [123, 456, 789],
                "lat": [1, 3, 5],
                "lon": [10, 20, 30],
            }
        ).set_index(["time", "id"])
        geometry = gpd.points_from_xy(df.lon, df.lat)
        gdf = gpd.GeoDataFrame(df, geometry=geometry)
        bbox = tp.space.BBox(max_lat=4, min_lat=0, max_lon=25, min_lon=0)
        results = tp.space.crop_to_bbox(gdf, bbox=bbox)[["lat", "lon"]].values
        expected = np.array([[1, 10], [3, 20]])
        np.testing.assert_equal(results, expected)

    def test_xarray_dataset(self):
        ds = xr.open_dataset(TEST_DATA_PATH / "lat_10_lon_10_time_10_D_test_data.nc")
        bbox = tp.space.BBox(min_lat=3.3, max_lat=5.7, min_lon=12.3, max_lon=14.4)
        results = tp.space.crop_to_bbox(ds, bbox=bbox)["var"].values
        expected = np.array(
            [
                [[0, 1, 2, 3, 4, 5, 6, 7, 8, 9], [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]],
                [[0, 1, 2, 3, 4, 5, 6, 7, 8, 9], [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]],
            ]
        )
        np.testing.assert_almost_equal(results, expected)


class TestPointsInRadius(TestCase):

    df = pd.read_csv(TEST_DATA_PATH / "station_metadata.csv")

    def test_points_and_distance(self):
        result = tp.space.points_within_radius(self.df, point=[43.5, 1.2], radius_km=10, return_distance=True)
        np.testing.assert_almost_equal(
            result[["id", "distance_km"]].values,
            np.array([[7628099999, 6.8720606], [7629399999, 7.6216178]], dtype=object),
        )


class TestGenerateGrid(TestCase):
    def test_filled_dataset(self):
        ds = tp.space.generate_grid(fill_value=1, return_dataset=True, resolution=0.5)
        result = ", ".join(f"{i} {j}" for i, j in ds.dims.items())
        expected = "lat 361, lon 721"
        self.assertEqual(result, expected)
        self.assertEqual(ds["var"].isel(lat=10, lon=10).item(), 1)

    def test_dataarray_from_bbox(self):
        da = tp.space.generate_grid(bbox=tp.space.BBox(min_lat=41, max_lat=52, min_lon=-5.5, max_lon=9.5))
        np.testing.assert_almost_equal(
            da.lat.values, np.array([41.0, 42.0, 43.0, 44.0, 45.0, 46.0, 47.0, 48.0, 49.0, 50.0, 51.0, 52.0])
        )
        np.testing.assert_almost_equal(
            da.lon.values,
            np.array([-5.5, -4.5, -3.5, -2.5, -1.5, -0.5, 0.5, 1.5, 2.5, 3.5, 4.5, 5.5, 6.5, 7.5, 8.5, 9.5]),
        )


class TestBBoxFromData(TestCase):

    ds = xr.open_dataset(TEST_DATA_PATH / "lat_10_lon_10_time_10_D_test_data.nc")

    def test_dataset(self):
        bbox = tp.space.bbox_from_data(self.ds)
        self.assertEqual(bbox.NWSE, (9, 11, 0, 20))

    def test_dataframe(self):
        df = self.ds.to_dataframe()
        bbox = tp.space.bbox_from_data(df)
        self.assertEqual(bbox.NWSE, (9, 11, 0, 20))

    def test_series(self):
        series = self.ds.to_dataframe()["var"]
        bbox = tp.space.bbox_from_data(series)
        self.assertEqual(bbox.NWSE, (9, 11, 0, 20))


class TestMatchDataBBox(TestCase):

    input = xr.open_dataset(TEST_DATA_PATH / "lat_10_lon_10_time_10_D_test_data.nc")
    reference = input.isel(lat=slice(3, 7), lon=slice(3, 7))

    def test_dataset_and_dataset(self):
        result = tp.space.match_data_bbox(self.reference, self.input)
        np.testing.assert_equal(result.lat.values, np.array([3, 4, 5, 6]))
        np.testing.assert_equal(result.lon.values, np.array([14, 15, 16, 17]))

    def test_dataframe_and_dataset(self):
        df = self.reference.to_dataframe()
        df_new = tp.space.match_data_bbox(df, self.input)
        np.testing.assert_equal(df_new.lat.values, np.array([3, 4, 5, 6]))
        np.testing.assert_equal(df_new.lon.values, np.array([14, 15, 16, 17]))

        df = df.reset_index()
        df_new = tp.space.match_data_bbox(df, self.input)
        np.testing.assert_equal(df_new.lat.values, np.array([3, 4, 5, 6]))
        np.testing.assert_equal(df_new.lon.values, np.array([14, 15, 16, 17]))


class TestGetNearestPoint(TestCase):

    df = pd.read_csv(TEST_DATA_PATH / "station_metadata.csv")

    def test_single_nearest_point_geodesic_no_distance(self):
        result = tp.space.get_nearest_point(self.df, lats=43.44, lons=1.23, return_distance=False)
        np.testing.assert_equal(result[["lat", "lon"]].values, np.array([[43.45, 1.25]]))
        self.assertFalse("distance_km" in result.columns)

    def test_multiple_nearest_points_geodesic(self):
        result = tp.space.get_nearest_point(self.df, lats=[43.44, 42.1], lons=[1.23, 0.3])
        np.testing.assert_equal(result[["lat", "lon"]].values, np.array([[43.45, 1.25], [43.008, 1.103]]))
        np.testing.assert_almost_equal(result["distance_km"].values, np.array([1.96354701, 120.50944031]))

    def test_single_nearest_point_kdtree(self):
        result = tp.space.get_nearest_point(self.df, lats=43.44, lons=1.23, method="kdtree")
        np.testing.assert_equal(result[["lat", "lon"]].values, np.array([[43.45, 1.25]]))

    def test_multiple_nearest_points_kdtree(self):
        result = tp.space.get_nearest_point(self.df, lats=[43.44, 42.1], lons=[1.23, 0.3], method="kdtree")
        np.testing.assert_equal(result[["lat", "lon"]].values, np.array([[43.45, 1.25], [43.008, 1.103]]))


class TestGroupPointsByGrid(TestCase):

    df = pd.read_csv(TEST_DATA_PATH / "station_metadata.csv")

    def test_groups_with_cellsize(self):
        groups, lat_bounds, lon_bounds = tp.space.group_points_by_grid(self.df, cellsize=0.5)
        np.testing.assert_equal(
            groups.values,
            np.array(
                [list([7627099999]), list([7628099999, 7629399999]), list([7630099999, 7631099999])],
                dtype=object,
            ),
        )
        np.testing.assert_equal(lat_bounds, np.array([42.758, 43.258, 43.758, 44.258]))
        np.testing.assert_equal(lon_bounds, np.array([0.853, 1.353, 1.853]))

    def test_groups_with_arbitrary_bounds(self):
        lat_bounds = [42.5, 43.4, 43.6, 44]
        lon_bounds = [1, 1.3, 1.4]
        # cellsize will be ignored
        groups = tp.space.group_points_by_grid(
            self.df, cellsize=0.5, lat_bounds=lat_bounds, lon_bounds=lon_bounds, return_cell_bounds=False
        )
        np.testing.assert_equal(
            groups.values,
            np.array(
                [
                    list([7627099999]),
                    list([7628099999, 7629399999]),
                    list([7631099999]),
                    list([7630099999]),
                ],
                dtype=object,
            ),
        )



"""
SPI: Number of rolling months depends on the timescale of interest, where 3 month SPI
is used for a short-term or seasonal drought index, 12 month SPI for an
intermediate-term drought index, and 48 month SPI for a long-term drought index.

TODO:
- An addition parameter should be the output timestep, as e.g. input data could be daily, but we
want data with monthly timesteps. A resampling operation should be carried out PRIOR to the rolling operation.
"""
import typing as T

import numpy as np
import pandas as pd
import scipy.stats as st
import xarray as xr

import terrapyn as tp

# TODO - this could be implemented in fit Gamma PDF if there is also the option to set a min_threshold
# just above zero, instead of equal to zero, e.g.
# prob_zero = _fraction_less_equal_threshold(values, 0.0001)

# def _fraction_less_equal_threshold(values: np.ndarray = None, threshold: float = 0) -> float:
#     return np.sum(values <= threshold) / values.shape[0]


# def _ensure_monthly_timesteps(data):
#     times = tp.time.get_time_from_data(data, time_dim=time_dim)
#     inferred_freq = pd.infer_freq(times)
#     # return inferred_freq[0] == 'M'

#     # Check if frequency of data is higher than monthly
#     if tp.time.FREQUENCY_DICT[inferred_freq[0]] < tp.time.FREQUENCY_DICT['M']:

#         # Resample to monthly

#     inferred_freq


# def _infer_frequency(data, time_dim: str = "time"):
#     """Return the frequency of data as a np.timedelta64"""
#     times = tp.time.get_time_from_data(data, time_dim=time_dim)
#     return np.diff(timests)).mean()


def _fit_gamma_pdf(array: np.ndarray = None) -> np.ndarray:
    """
    Fits a Gamma PDF to a 1-D array and returns the shape and scale parameters,
    where shape = alpha and scale = 1 / beta see
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.gamma.html.

    Values < 0 are automatically clipped to 0, and only finite values > 0 are included in the fit.
    If only non-finite values are given, a parameter array of NaN is returned.

    Args:
        array: Array of values to fit

    Returns:
        Array of [shape, scale] parameters for fitted Gamma PDF
    """
    # Include only finite values > 0
    finite_values_mask = np.isfinite(array) & (array > 0)

    if np.any(finite_values_mask):

        finite_values = array[finite_values_mask]

        # # Fit Gamma PDF to data using scipy
        # shape, _, scale = st.gamma.fit(finite_values)
        # return np.array([shape, scale])

        # Fit Gamma PDF using approximation - see Lloyd-Hughes and Saunders 2002, A drought climatology for Europe
        log_values = np.log(finite_values)
        mean_values = np.mean(finite_values)
        A = np.log(mean_values) - np.mean(log_values)
        alpha = 1 / (4 * A) * (1 + np.sqrt(1 + 4 / 3 * A))
        beta = mean_values / alpha
        return np.array([alpha, beta])

    else:
        return np.array([np.nan, np.nan])


def _fit_gamma_pdf_dataarray(da: xr.DataArray = None, time_dim: str = "time") -> xr.Dataset:
    """
    Fits a Gamma PDF to an xr.DataArray and returns an xr.Dataset with the Gamma function shape and scale parameters,
    where shape = alpha and scale = 1 / beta see
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.gamma.html.

    Values < 0 are automatically clipped to 0, and only finite values are included in the fit.

    Args:
        da: xr.DataArray of values to fit, with coordinates of latitude, longitude, and time.
        time_dim: Name of the time coordinate in `da`.

    Returns:
        xr.Dataset of the Gamma function shape and scale parameters
    """
    # Fit the Gamma function to all values in `time_dim`, for each coordinate
    parameters = xr.apply_ufunc(
        _fit_gamma_pdf,
        da,
        input_core_dims=[[time_dim]],
        output_core_dims=[["parameter"]],
        vectorize=True,
        dask="parallelized",
        output_dtypes=["float32"],
        dask_gufunc_kwargs=dict(output_sizes={"parameter": 2}),
    )

    # Convert dataarray to dataset with labelled variables
    parameters = parameters.to_dataset(dim="parameter").rename({0: "shape", 1: "scale"})
    return parameters


def _calc_gamma_cdf(
    array: T.Union[pd.Series, np.ndarray] = None, shape: float = None, scale: float = None
) -> T.Union[pd.Series, np.ndarray]:
    """
    Returns the Gamma CDF for values in a 1-D array, for the given shape and scale parameters,
    where shape = alpha and scale = 1 / beta
    see https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.gamma.html

    Scales the Gamma CDF by the fraction of zero values in the data - see "A drought climatology
    for Europe, B. Lloyd-Hughes and M.A. Saunders 2002, https://doi.org/10.1002/joc.846

    Args:
        array: Data values to use for generating the CDF.
        shape: The shape (alpha) parameter for the Gamma function.
        scale: The scale (1 / beta) parameter for the Gamma function.

    Returns:
        Array of values for the Gamma function CDF
    """
    # Probabity of zeros
    # prob_zero = _fraction_less_equal_threshold(array, min_threshold)
    prob_zero = np.sum(array == 0.0) / array.shape[0]

    # Calculate the gamma CDF of the data, fixing location to zero
    gamma_cdf = prob_zero + (1 - prob_zero) * st.gamma.cdf(array, a=shape, loc=0, scale=scale)

    if isinstance(array, pd.Series):
        return pd.Series(gamma_cdf, index=array.index)
    else:
        return gamma_cdf


def _calc_gamma_cdf_dataarray(
    da_values: xr.DataArray = None,
    da_parameters: xr.Dataset = None,
    time_dim: str = "time",
    shape_dim: str = "shape",
    scale_dim: str = "scale",
) -> xr.DataArray:
    """
    Calculates a Gamma CDF for valus at each coordinate in a xr.DataArray using the shape and scale parameters,
    where shape = alpha and scale = 1 / beta see
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.gamma.html.

    Args:
        da_values: xr.DataArray of values, with coordinates of latitude, longitude, and time.
        da_parameters: xr.Dataset of Gamma function parameters "shape" and "scale"
        time_dim: Name of the time coordinate in `da_values`.
        shape_dim: Name of the Gamma "shape" parameter in `da_parameters`.
        scale_dim: Name of the Gamma "scale" parameter in `da_parameters`.

    Returns:
        xr.DataArray for the Gamma function CDF for the given values, shape and scale parameters.
    """
    # Fit the Gamma function to all values in `time_dim`, for each coordinate
    return xr.apply_ufunc(
        _calc_gamma_cdf,
        da_values,
        da_parameters[shape_dim],
        da_parameters[scale_dim],
        input_core_dims=[[time_dim], [], []],
        output_core_dims=[[time_dim]],
        vectorize=True,
        dask="parallelized",
        output_dtypes=["float32"],
    )


def fit_gamma_pdf(
    data: T.Union[pd.Series, xr.DataArray] = None, time_dim: str = "time"
) -> T.Union[T.Tuple, xr.DataArray]:
    """
    Fits a Gamma PDF to data and return the shape and scale parameters of the Gamma function,
    where shape = alpha and scale = 1 / beta see
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.gamma.html.

    Normally data is a rolling sum for monthly data.

    Accept pd.Series or xr.DataArray, where for pd.Series a tuple of (shape, scale) is returned, and
    for xr.DataArray a xr.Dataset is returned with shape and scale as variables.

    Values < 0 are automatically clipped to 0, and only finite values are included in the fit.

    Args:
        data: Data to fit
        time_dim: (Optional) Name of the time coordinate in `data`. Only applies if using a xr.DataArray.

    Returns:
        Shape and scale parameters for the fitted Gamma PDF
    """
    if not isinstance(data, (pd.Series, xr.DataArray)):
        raise TypeError("data must be of type pd.Series or xr.DataArray")

    # Set negative values to zero
    data = data.clip(0, None)

    if isinstance(data, pd.Series):
        return _fit_gamma_pdf(data)

    else:
        if tp.dask_utils.uses_dask(data):
            # Re-chunk along time_dim if using Dask
            data = tp.dask_utils.chunk_xarray(data, coords_no_chunking=time_dim)

        return _fit_gamma_pdf_dataarray(da=data, time_dim=time_dim)


def calc_gamma_cdf(
    data: T.Union[pd.Series, xr.DataArray] = None,
    gamma_parameters: T.Union[T.Tuple, xr.Dataset] = None,
    time_dim: str = "time",
    shape_dim: str = "shape",
    scale_dim: str = "scale",
) -> T.Union[T.Tuple, xr.DataArray]:
    """
    Calculates a Gamma CDF for valus at each coordinate in a pd.Series or xr.DataArray using
    the shape and scale parameters, where shape = alpha and scale = 1 / beta see
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.gamma.html.

    Args:
        data: pd.Series or xr.DataArray of values. xr.DataArray should have coordinates of latitude,
        longitude, and time (`time_dim').
        gamma_parameters: Gamma function parameters, either a Tuple of ("shape", "scale") or xr.Dataset
        with variables of "shape", "scale".
        time_dim: Name of the time coordinate in `gamma_parameters`. Only applies to xr.Dataset.
        shape_dim: Name of the Gamma "shape" parameter in `gamma_parameters`. Only applies to xr.Dataset.
        scale_dim: Name of the Gamma "scale" parameter in `gamma_parameters`. Only applies to xr.Dataset.

    Returns:
        Parameters of the Gamma function CDF for the given values (shape and scale parameters).
    """
    if not isinstance(data, (pd.Series, xr.DataArray)):
        raise TypeError("data must be of type pd.Series or xr.DataArray")

    if isinstance(data, pd.Series):
        return _calc_gamma_cdf(data, shape=gamma_parameters[0], scale=gamma_parameters[1])

    else:
        if tp.dask_utils.uses_dask(data):
            # Re-chunk along time_dim if using Dask
            data = tp.dask_utils.chunk_xarray(data, coords_no_chunking=time_dim)

        return _calc_gamma_cdf_dataarray(
            data, gamma_parameters, time_dim=time_dim, shape_dim=shape_dim, scale_dim=scale_dim
        )


def _cdf_to_normal_ppf(cdf: np.ndarray = None) -> np.ndarray:
    """Apply the inverse normal distribution to a CDF to yield a normal
    Percent point function (PPF) with mean = 0 and std = 1
    """
    normal_ppf = st.norm.ppf(cdf, loc=0, scale=1)
    normal_ppf[np.isinf(normal_ppf)] = np.nan
    return normal_ppf


def _cdf_to_normal_ppf_dataarray(da: xr.DataArray = None, time_dim: str = "time") -> xr.DataArray:
    """Apply the inverse normal distribution to a CDF to yield a normal PPF with mean = 0 and std = 1"""
    return xr.apply_ufunc(
        _cdf_to_normal_ppf,
        da,
        input_core_dims=[[time_dim]],
        output_core_dims=[[time_dim]],
        vectorize=True,
        dask="parallelized",
        output_dtypes=["float32"],
    )


def cdf_to_normal_ppf(
    data: T.Union[pd.Series, xr.DataArray] = None, time_dim: str = "time"
) -> T.Union[pd.Series, xr.DataArray]:
    """
    Apply the inverse normal distribution to a CDF to yield a normal PPF with mean = 0 and std = 1

    Args:
        data: pd.Series or xr.DataArray of values. xr.DataArray should have coordinates of latitude,
        longitude, and time (`time_dim').
        time_dim: Name of the time coordinate in `data`. Only applies to xr.DataArray.

    Returns:
        The computed values of the inverse normal distribution for the given CDF
    """
    if not isinstance(data, (pd.Series, xr.DataArray)):
        raise TypeError("data must be of type pd.Series or xr.DataArray")

    if isinstance(data, pd.Series):
        normal_pdf = _cdf_to_normal_ppf(data)
        return pd.Series(normal_pdf, index=data.index)

    else:
        if tp.dask_utils.uses_dask(data):
            # Re-chunk along time_dim if using Dask
            data = tp.dask_utils.chunk_xarray(data, coords_no_chunking=time_dim)
        return _cdf_to_normal_ppf_dataarray(data, time_dim=time_dim)


def calc_spi(
    data: T.Union[pd.Series, xr.DataArray] = None,
    n_months: int = 3,
    gamma_parameters: T.Union[pd.Series, xr.DataArray] = None,
    return_gamma_params: bool = False,
    time_dim: str = "time",
) -> T.Union[pd.Series, xr.DataArray]:
    """Calculate SPI, where `n_months` is the timescale of interest: 3 month SPI is used for a
    short-term or seasonal drought index, 12 month SPI for an intermediate-term drought index,
    and 48 month SPI for a long-term drought index.

    Groups of monthly data are modelled using a Gamma Distritibution, and the corresponding
    CDF of model is transformed to a Normal PPF for each monthly value, yielding the SPI value.

    Args:
        data: Data, with values on a monthly timestep with datetimes as `time_dim` index
        n_months: The number of months over which to calculate the rolling mean (the timescale of interest)
        gamma_parameters: (Optional) The shape and scale parameters the define a Gamma distribution. If given,
        these parameters are used instead of fitting the data. For `pd.Series` the index must be the month,
        with values of a Tuple of (shape, scale).
        return_gamma_params: If `True` then return fitted Gamma Parameters.
        time_dim: Name of the time coordinate in `data`. Only applies to xr.DataArray.

    Returns:
        SPI values for each month
    """
    # Calculate rolling mean over `n_months` with minimum of `n_months` in the window
    data_rolling = tp.time.rolling(data, n_periods=n_months, min_periods=n_months, method="mean", time_dim=time_dim)

    # Group rolling data by month
    data_rolling_grouped = tp.time.groupby_time(data_rolling, grouping="month")

    # Use Gamma Parameters provided, otherwise fit the Gamma distribution
    if gamma_parameters is None:

        # Calculate Gamma distribution parameters for each monthly group
        gamma_parameters = data_rolling_grouped.apply(tp.indices.spi.fit_gamma_pdf)

    # For each month, calculate the Gamma CDF
    if isinstance(data, pd.Series):

        # Check length of gamma_parameters is equal to the number of months
        # gamma_parameters is a list of tuples of (shape, scale)
        # if len(gamma_parameters) != data_rolling_grouped.ngroups:
        #     raise ValueError(f"`gamma_parameters` has length {len(gamma_parameters)}")

        gamma_cdf_list = [
            tp.indices.spi.calc_gamma_cdf(group, gamma_parameters.loc[label]) for label, group in data_rolling_grouped
        ]

        # For each month, transform the Gamma CDF to a Normal PPF, then sort the data by time
        normal_ppf = pd.concat([tp.indices.spi.cdf_to_normal_ppf(item) for item in gamma_cdf_list]).sort_index()

    else:
        gamma_cdf_list = [
            tp.indices.spi.calc_gamma_cdf(group, gamma_parameters.sel({"month": label}).drop("month"))
            for label, group in data_rolling_grouped
        ]

        # For each month, transform the Gamma CDF to a Normal PPF, then sort the data by time
        normal_ppf = xr.concat([tp.indices.spi.cdf_to_normal_ppf(item) for item in gamma_cdf_list], dim="time").sortby(
            "time"
        )

    if return_gamma_params:
        return normal_ppf, gamma_parameters
    else:
        return normal_ppf



import typing as T

import dask
import xarray as xr

import terrapyn as tp


def create_cluster_and_client(n_workers: int = 4, threads_per_worker: int = 1, memory_limit: str = "1GB", **kwargs):
    """Create a local cluster and initiate a client

    Args:
        n_workers: Number of workers to start.
        threads_per_worker: Number of threads per each worker.
        memory_limit: Memory limit per worker.

    Returns:
        Dask Client
    """
    client = dask.distributed.Client(
        n_workers=n_workers, threads_per_worker=threads_per_worker, memory_limit=memory_limit, **kwargs
    )
    print(f"Client dashboard: {client.dashboard_link}")
    return client


def chunk_xarray(
    data: T.Union[xr.Dataset, xr.DataArray] = None,
    coords_no_chunking: T.Union[str, T.Iterable[str]] = None,
    coords_chunking: T.Dict = None,
) -> T.Union[xr.Dataset, xr.DataArray]:
    """
    Chunks xarrary data structures. If coordinate names are not given in `coords_no_chunking`, their chunk sizes
    are automatically determined by Dask. Otherwise, they can be set explicitly with `coords_chunking`

    Args:
        coords_no_chunking: List of coordinates that will have no chunking along this direction
        coords_chunking: (Optional) Dictionary of {coordinate name: chunk size} that will set the
        chunk size for those coordinates.

    Returns:
        Xarray data that has been chunked
    """
    # List of all coords in data
    coords = list(data.coords)

    # create chunks dict
    chunks = {}

    if coords_no_chunking is not None:
        # ensure coords_no_chunking is a list
        coords_no_chunking = tp.utils.ensure_list(coords_no_chunking)

        # Generate dict with coords that will not be chunked, where -1 means no chunking along this dimension
        chunks.update({i: -1 for i in coords_no_chunking})

        # Remove non-chunked coords from list of coords
        coords = tp.utils.remove_list_elements(coords, coords_no_chunking)

    if coords_chunking is not None:
        # combine provided chunk sizes with existing chunks
        chunks.update(coords_chunking)

        # Remove these coords from list of coords
        coords = tp.utils.remove_list_elements(coords, coords_chunking.keys())

    # Finally, set chunk sizes to 'auto' for all remaining coords
    if len(coords) > 0:
        chunks.update({i: "auto" for i in coords})

    return data.chunk(chunks)


def uses_dask(data: T.Union[xr.Dataset, xr.DataArray] = None) -> T.Union[xr.Dataset, xr.DataArray]:
    """Check if an xarray data structure uses Dask"""
    if isinstance(data, xr.DataArray) and isinstance(data.data, dask.array.Array):
        return True
    if isinstance(data, xr.Dataset) and any(isinstance(var.data, dask.array.Array) for var in data.variables.values()):
        return True
    return False



import datetime as dt
import typing as T
import zoneinfo

import numpy as np
import pandas as pd
import xarray as xr
from dateutil import rrule
from dateutil.relativedelta import relativedelta

import terrapyn as tp

# import dateutil


def datetime64_to_datetime(datetime64: np.datetime64 = None) -> dt.datetime:
    """Convert numpy.datetime64 to dt.datetime"""
    return np.datetime64(datetime64, "us").astype(dt.datetime)


def datetime_to_datetime64(time: dt.datetime = None) -> np.datetime64:
    """Convert dt.datetime to numpy.datetime64"""
    return np.datetime64(time)


def time_offset(
    time: dt.datetime = None,
    years: int = 0,
    months: int = 0,
    weeks: int = 0,
    days: int = 0,
    hours: int = 0,
    minutes: int = 0,
    seconds: int = 0,
    microseconds: int = 0,
    **kwargs,
) -> dt.datetime:
    """
    Apply a time offset to a datetime, using calendar months, taking care of leap years.
    Accepts positive or negative offset values.

    Args:
        time: The input datetime
        years: Number of calendar years
        months: Number of calendar months
        weeks: Number of calendar weeks
        days: Number of days
        hours: Number of hours
        minutes: Number of minutes
        seconds: Number of seconds
        microseconds: Number of microseconds
        kwargs: kwargs to pass to dateutils.relativedelta.relativedelta

    Example 1: Positive monthly offset
        >>> time_offset(dt.datetime(1994, 11, 5), months=1)
        datetime.datetime(1994, 12, 5, 0, 0)

    Example 2: Negative monthly offset
        >>> time_offset(dt.datetime(1994, 11, 5), months=-1)
        datetime.datetime(1994, 10, 5, 0, 0)

    Example 3: 1 day previous
        >>> time_offset(dt.datetime(1994, 11, 5), days=-1)
        datetime.datetime(1994, 11, 4, 0, 0)

    Example 4: This time on previous day
        >>> time_offset(dt.datetime(1994, 11, 5, 7, 23), days=-1)
        datetime.datetime(1994, 11, 4, 7, 23)

    Example 5: 6 hours previous
        >>> time_offset(dt.datetime(1994, 11, 5, 7, 23), hours=-6)
        datetime.datetime(1994, 11, 5, 1, 23)

    Example 6: 1 day and 18 hours previous
        >>> time_offset(dt.datetime(1994, 11, 5, 7, 23), days=-1, hours=-18)
        datetime.datetime(1994, 11, 3, 13, 23)

    Example 7: In 27 hours time
        >>> time_offset(dt.datetime(1994, 11, 5, 7, 23), hours=27)
        datetime.datetime(1994, 11, 6, 10, 23)
    """
    return time + relativedelta(
        years=years,
        months=months,
        weeks=weeks,
        days=days,
        hours=hours,
        minutes=minutes,
        seconds=seconds,
        microseconds=microseconds,
        **kwargs,
    )


def get_time_from_data(
    data: T.Union[
        xr.Dataset,
        xr.DataArray,
        pd.DataFrame,
        pd.Series,
        np.ndarray,
        T.List,
        dt.datetime,
        pd.DatetimeIndex,
        pd.MultiIndex,
    ] = None,
    time_dim: str = "time",
) -> pd.DatetimeIndex:
    """
    Returns a pd.DatetimeIndex for the values of `time_dim` in the data.

    Args:
        data: Input data which contains datetime-like objects, optionally in the column/dimension `time_dim`
        time_dim: Name of dimension that has dt.datetime objects (if required, default==`'time'`).
        Ignored for pd.Series, np.ndarray and list types.

    Returns:
        pd.DatetimeIndex of times in data
    """
    # Perform data type checks and convert data to a pandas DatetimeIndex
    if isinstance(data, (pd.Series, pd.DataFrame)):
        # Check if `time_dim` is the index
        if time_dim in data.index.names:
            times = data.index.get_level_values(time_dim)
        else:
            if isinstance(data, pd.Series):
                times = pd.DatetimeIndex(data)
            else:
                times = pd.DatetimeIndex(data[time_dim])
    elif isinstance(data, (xr.Dataset, xr.DataArray)):
        times = data.indexes[time_dim]
    elif isinstance(data, (np.ndarray, list)):
        times = pd.DatetimeIndex(data, name="time")
    elif isinstance(data, dt.datetime):
        times = pd.DatetimeIndex([data], name="time")
    elif isinstance(data, pd.DatetimeIndex):
        times = data
    elif isinstance(data, pd.MultiIndex):
        # Check if `time_dim` is the index
        if time_dim in data.names:
            times = data.get_level_values(time_dim)
        else:
            raise ValueError(f"`time_dim` of {time_dim} not found in pd.Multindex")
    else:
        data_types_str = ", ".join(
            str(i)
            for i in [
                xr.Dataset,
                xr.DataArray,
                pd.DataFrame,
                pd.Series,
                np.ndarray,
                T.List,
                dt.datetime,
                pd.DatetimeIndex,
            ]
        )
        raise TypeError(f"Data is of type {type(data)} but must be one of type: {data_types_str}")
    return times


def groupby_time(
    data: T.Union[xr.Dataset, xr.DataArray, pd.DataFrame, pd.Series] = None,
    time_dim: str = "time",
    grouping: str = "week",
    other_grouping_keys: T.Optional[T.Union[str, T.List[str]]] = None,
):
    """
    Generate a `groupby` object where data are grouped by `time` and (optionally) `id`.
    Works with Pandas Series/DataFrame as well as Xarray Dataset/DataArray.

    Args:
        data: Input data.
        time_dim: Name of index/column/dimension containing datetime-like objects.
        grouping: Time range of grouping, one of 'week', 'month', 'dayofyear', 'dekad', 'pentad'.
        other_grouping_keys: (Optional, only for pd.Dataframe/pd.Series) Other keys to use to group
        the data, in addition to the time grouping.

    Returns:
        Groupby object (pandas or xarray type)
    """
    # Extract/Convert the `time_dim` to a pandas.DatetimeIndex
    times = get_time_from_data(data, time_dim=time_dim)

    if grouping == "week":
        time_groups = times.isocalendar().week
    elif grouping == "month":
        time_groups = times.month.values
    elif grouping == "year":
        time_groups = times.year.values
    elif grouping == "dayofyear":
        time_groups = get_day_of_year(times, time_dim, modify_ordinal_days=True)
    elif grouping == "dekad":
        time_groups = datetime_to_dekad_number(times)
    elif grouping == "pentad":
        time_groups = datetime_to_pentad_number(times)
    else:
        raise ValueError("`grouping` must be one of 'dayofyear', 'week', 'month', 'year', 'pentad', 'dekad'")

    if isinstance(time_groups, pd.Series):
        time_groups = time_groups.values

    if isinstance(data, (xr.DataArray, xr.Dataset)):
        time_groups = xr.DataArray(time_groups, dims=time_dim, name=grouping)
    else:
        time_groups = pd.Index(data=time_groups, name=grouping)

    # Optionally group by other keys as well as `time_dim`
    if other_grouping_keys is not None:
        if isinstance(other_grouping_keys, str):
            other_grouping_keys = [other_grouping_keys]
        grouper = [time_groups] + other_grouping_keys
    else:
        grouper = time_groups

    return data.groupby(grouper)


def get_day_of_year(
    data: T.Union[
        xr.Dataset,
        xr.DataArray,
        pd.DataFrame,
        pd.Series,
        np.ndarray,
        T.List,
        dt.datetime,
        pd.DatetimeIndex,
    ] = None,
    time_dim: str = "time",
    modify_ordinal_days: bool = False,
) -> np.ndarray:
    """
    Returns array of day of year based on the times in `data`. If `modify_ordinal_days` is `True`,
    the days are modified so if a year has 366 days, both 29 February and 1 March are assigned 60.
    The returned array could be used as a 'dayofyear' variable in an xr.Dataset or pd.DataFrame
    to assist with groupby operations, fixing the issue with leap years and
    standard dt.dayofyear functionality.

    Args:
        data: Input data which contains datetime objects in `time_dim`
        time_dim: Name of dimension that has dt.datetime objects (if required, default='time').
        Ignored for pd.Series, np.ndarray and list types.
        modify_ordinal_days: If `True`, then if a year has 366 days, both 29 Feb and 1 Mar are
        assigned day of year 60.

    Returns:
        Array of 'days of year'
    """
    times = get_time_from_data(data, time_dim=time_dim)
    if modify_ordinal_days:
        march_or_later = (times.month >= 3) & times.is_leap_year
        ordinal_day = times.dayofyear.values
        modified_ordinal_day = ordinal_day
        modified_ordinal_day[march_or_later] -= 1
        return modified_ordinal_day
    else:
        return times.dayofyear.values


def datetime_to_pentad_number(
    times: T.Union[pd.DatetimeIndex, pd.Series, dt.datetime, T.List, np.ndarray] = None
) -> np.ndarray:
    """
    Determine pentad number from a datetime object, where a pentad is a group
    of 5 days, with 73 pentads in a year. Works for standard years (365 days) and
    leap years (366 days). Accepts single dt.datetime or a list/array of dt.datetime
    objects.

    Args:
        times: Datetime(s)

    Returns:
        Pentad number for the date(s)
    """
    days_of_year = get_day_of_year(times, modify_ordinal_days=True)
    pentads = np.arange(1, 366, 5)
    return np.digitize(days_of_year, bins=pentads, right=False)


def datetime_to_dekad_number(
    dates: T.Union[pd.DatetimeIndex, pd.Series, dt.datetime, np.datetime64, T.List, np.ndarray] = None
) -> np.ndarray:
    """
    Determine dekad number from a datetime object, where a dekad is a group of 10 days, with 36 dekads in a year.
    Works for standard (365 days) and leap (366 days) years. Accepts single dt.datetime or a list/array
    of dt.datetime objects.

    Args:
        dates: Date(s) to use

    Returns:
        Array of dekad number for the given dates

    Example:
        >>> datetime_to_dekad_number(np.array([np.datetime64('2004-01-01'), np.datetime64('2004-02-05')]))
        array([1, 4])
        >>> datetime_to_dekad_number([dt.datetime(2004, 1, 1), dt.datetime(2004, 1, 11), dt.datetime(2004, 2, 5)])
        array([1, 2, 4])
        >>> datetime_to_dekad_number(dt.datetime(2004, 2, 5))
        array([4])

    """
    if isinstance(dates, (dt.datetime, np.datetime64)):
        dates = [dates]
    dates = pd.DatetimeIndex(dates)
    count = np.digitize(dates.day, bins=[10, 20], right=True) + 1  # Add 1 since bins start at zero
    dekads = (dates.month - 1) * 3 + count
    return dekads.values


def daily_date_range(
    start_time: dt.datetime = None,
    end_time: dt.datetime = None,
    delta_days: int = None,
    hours: T.Union[int, T.List[int]] = None,
    reset_time: bool = True,
    ref_hour: int = 0,
    ref_minutes: int = 0,
    ref_seconds: int = 0,
    ref_microseconds: int = 0,
) -> T.List[dt.datetime]:
    """
    Generate a list of dates with a daily frequency, where a datetime objects is generated for each
    given hour in `hours`. If `start_time` or `end_time` is `None` today's date is used.
    If `delta_days` is given, a range of dates are generated using this day offset and the `start_time`.
    Otherwise, the range is between `start_time` and `end_time`. If `reset_time==True`, the hours, minutes,
    seconds and microseconds are replaced with the given reference values `ref_hour` etc.

    Args:
        start_time: Start date for range
        end_time: End date for range
        delta_days: The number of days to offset the `start_time`, where the generated date range
        is from this offset date to the `start_time`. Can be positive or negative.
        hours: The hours that will be generated for each day, such that multiple datetimes will be
        generated with the same day value, and where the hour are those given in the `hours` lsit
        reset_time: If `True`, the hours, minutes, seconds and microseconds are replaced with
        the given reference values `ref_hour` etc. This should be `True` if you want include all
        days, even if the day is not complete

    Returns:
        List of dates

    Example 1: Daily date range with hours, minutes and seconds reset to 0

        >>> start_time = dt.datetime(1994, 11, 5, 7, 23)
        >>> end_time = dt.datetime(1994, 11, 7, 0)
        >>> daily_date_range(start_time, end_time)  # doctest: +NORMALIZE_WHITESPACE
        [datetime.datetime(1994, 11, 5, 0, 0),
         datetime.datetime(1994, 11, 6, 0, 0),
         datetime.datetime(1994, 11, 7, 0, 0)]

    Example 2: Daily date range with a specific hour

        >>> start_time = dt.datetime(1994, 11, 5, 7, 23)
        >>> end_time = dt.datetime(1994, 11, 7, 0)
        >>> daily_date_range(start_time, end_time, hours=3)  # doctest: +NORMALIZE_WHITESPACE
        [datetime.datetime(1994, 11, 5, 3, 0),
         datetime.datetime(1994, 11, 6, 3, 0),
         datetime.datetime(1994, 11, 7, 3, 0)]

    Example 3: Daily date range with multiple hours

        >>> start_time = dt.datetime(1994, 11, 5, 7, 23)
        >>> end_time = dt.datetime(1994, 11, 7, 0)
        >>> daily_date_range(start_time, end_time, hours=[3, 6])  # doctest: +NORMALIZE_WHITESPACE
        [datetime.datetime(1994, 11, 5, 3, 0),
         datetime.datetime(1994, 11, 5, 6, 0),
         datetime.datetime(1994, 11, 6, 3, 0),
         datetime.datetime(1994, 11, 6, 6, 0),
         datetime.datetime(1994, 11, 7, 3, 0),
         datetime.datetime(1994, 11, 7, 6, 0)]

    Example 4: Delta days range, only taking the day into account

        >>> date = dt.datetime(1994, 11, 5, 7, 23)
        >>> daily_date_range(date, delta_days=-1, hours=[3, 9])  # doctest: +NORMALIZE_WHITESPACE
        [datetime.datetime(1994, 11, 4, 3, 0),
         datetime.datetime(1994, 11, 4, 9, 0),
         datetime.datetime(1994, 11, 5, 3, 0),
         datetime.datetime(1994, 11, 5, 9, 0)]

    Example 5: Delta days range, taking hours, minutes and seconds into account

        >>> date = dt.datetime(1994, 11, 5, 7, 23)
        >>> daily_date_range(date, delta_days=-1, reset_time=False, hours=[3, 9])  # doctest: +NORMALIZE_WHITESPACE
        [datetime.datetime(1994, 11, 4, 9, 23),
         datetime.datetime(1994, 11, 5, 3, 23),
         datetime.datetime(1994, 11, 5, 9, 23)]
    """
    date_today = dt.datetime.today()
    if start_time is None:
        start_time = date_today
    if end_time is None:
        end_time = date_today

    if reset_time:
        # Replace hours, minutesd, seconds and microseconds
        start_time = start_time.replace(
            hour=ref_hour,
            minute=ref_minutes,
            second=ref_seconds,
            microsecond=ref_microseconds,
        )
        end_time = end_time.replace(
            hour=ref_hour,
            minute=ref_minutes,
            second=ref_seconds,
            microsecond=ref_microseconds,
        )

    if delta_days is not None:
        delta_date = time_offset(start_time, days=delta_days)
        if delta_days < 0:
            end_time = start_time
            start_time = delta_date
        else:
            end_time = delta_date

    # If hours is given, set the hour of `end_time` to be the maximum hour
    if hours is not None:
        if isinstance(hours, int):
            max_hour = hours
        else:
            max_hour = max(hours)
        end_time = end_time.replace(hour=max_hour)

    return list(rrule.rrule(freq=rrule.DAILY, dtstart=start_time, until=end_time, byhour=hours))


def monthly_date_range(
    start_time: dt.datetime = None,
    end_time: dt.datetime = None,
    delta_months: int = None,
    reset_time: bool = True,
) -> T.List[dt.datetime]:
    """
    Generate a list of dates with a frequency of 1 month. If `start_time` is `None` today's date is used.
    If `delta_months` is given, a range of dates are generated using this monthly offset and the `start_time`.
    Otherwise, the range is between `start_time` and `end_time`. If `reset_time==True`, the days and time are
    ignored and only the year and month are used (with the days set to 1). This is useful if you want to include the
    first and last month even if they are not complete.

    Args:
        start_time: Start date for range
        end_time: End date for range
        delta_months: The number of calendar months to offset the `start_time`, where the
        generated date range is from this offset date to the `start_time`. Can be positive or negative.
        reset_days: Whether to ignore the days of the datetime, and reset the days to 1

    Returns:
        List of dates

    Example 1: Generate a range of dates with monthly frequency, ignoring the days (reset to 1)

        >>> start_time = dt.datetime(1994, 11, 5)
        >>> end_time = dt.datetime(1995, 3, 1)
        >>> monthly_date_range(start_time, end_time)  # doctest: +NORMALIZE_WHITESPACE
        [datetime.datetime(1994, 11, 1, 0, 0),
         datetime.datetime(1994, 12, 1, 0, 0),
         datetime.datetime(1995, 1, 1, 0, 0),
         datetime.datetime(1995, 2, 1, 0, 0),
         datetime.datetime(1995, 3, 1, 0, 0)]

    Example 2: Generate a range of dates with monthly frequency, where the day of the start date is retained

        >>> start_time = dt.datetime(1994, 11, 5)
        >>> end_time = dt.datetime(1995, 3, 1)
        >>> monthly_date_range(start_time, end_time, reset_time=False)  # doctest: +NORMALIZE_WHITESPACE
        [datetime.datetime(1994, 11, 5, 0, 0),
         datetime.datetime(1994, 12, 5, 0, 0),
         datetime.datetime(1995, 1, 5, 0, 0),
         datetime.datetime(1995, 2, 5, 0, 0)]

    Example 3: Use a delta months option to generate a range of dates with monthly frequency, starting from
    delta months before or after the start date
        >>> monthly_date_range(dt.datetime(1994, 11, 5), delta_months=1)
        [datetime.datetime(1994, 11, 1, 0, 0), datetime.datetime(1994, 12, 1, 0, 0)]

    """
    date_today = dt.datetime.today()
    if start_time is None:
        start_time = date_today
    if end_time is None:
        end_time = date_today

    if delta_months is not None:
        delta_date = time_offset(start_time, months=delta_months)
        if delta_months < 0:
            end_time = start_time
            start_time = delta_date
        else:
            end_time = delta_date

    if reset_time:
        start_time = dt.datetime(start_time.year, start_time.month, 1)
        end_time = dt.datetime(end_time.year, end_time.month, 1)

    return list(rrule.rrule(freq=rrule.MONTHLY, dtstart=start_time, until=end_time))


def add_day_of_year_variable(
    data: T.Union[xr.Dataset, xr.DataArray] = None, time_dim: str = "time", modify_ordinal_days: bool = True
) -> xr.Dataset:
    """
    Assign a day of year variable to an xr.dataset/dataarray that optionally
    accounts for years with 366 days by repeating day 60 (for February 29).
    Converts xr.DataArray to xr.Dataset so variable can be added.

    Args:
        data: Input dataset/dataarray
        time_dim: Name of time dimension
        modify_ordinal_days: If `True`, then if a year has 366 days, both 29 Feb
        and 1 Mar are assigned day of year 60.

    Returns:
        Dataset with additional dayofyear variable
    """
    day_of_year = get_day_of_year(data, time_dim, modify_ordinal_days=modify_ordinal_days)

    if isinstance(data, xr.DataArray):
        data = data.to_dataset()

    return data.assign({"dayofyear": (time_dim, day_of_year)})


def check_start_end_time_validity(
    start_time: T.Union[dt.datetime, np.datetime64, pd.Timestamp] = None,
    end_time: T.Union[dt.datetime, np.datetime64, pd.Timestamp] = None,
    verbose: bool = False,
) -> bool:
    """
    Check whether end date is after start date

    Args:
        start_time: Start date.
        end_time: End date.
        verbose: Whether to print a warning if end is before start.

    Returns:
        True if end date is after start date, else False

    Example:
        >>> check_start_end_time_validity(dt.datetime(2019, 2, 3), dt.datetime(2012, 1, 3), verbose=False)
        False

    """
    if start_time and end_time:
        if end_time < start_time:
            if verbose:
                print(f"Warning: End time {end_time} before start time {start_time}")
            return False
        else:
            return True
    else:
        raise ValueError("Both `start_time` and `end_time` must be given")


def list_timezones():
    """
    List all available timezones.
    """
    return zoneinfo.available_timezones()


def time_to_local_time(
    times: T.Union[dt.datetime, pd.Timestamp, pd.DatetimeIndex] = None, timezone_name: str = "UTC"
) -> T.Union[dt.datetime, pd.Timestamp, pd.DatetimeIndex]:
    """
    Apply a timezone / daylight-savings-time (dst) offset to a (naive) datetime object.
    The datetimes in `times` are assumed to be in UTC if there are timezone-naive
    (no `tzinfo` has been set). If a timezone is set, the appropriate offset is applied.

    Args:
        times: Datetimes where timezone-naive times are assumed to be in UTC, or the timezone is set.
        timezone_name: The name of the timezone, understood by `zoneinfo`.
        Use `time.list_timezones()` to view all possible options.

    Returns:
        Datetimes where the required offset has been applied to the time

    Example 1: Central Europe, day before daylight savings time (DST) change. Naive time in UTC.
        >>> date = dt.datetime(2020, 3, 28, 1, 15)
        >>> time_to_local_time(date, timezone_name='CET')
        Timestamp('2020-03-28 02:15:00')

    Example 2: Central Europe, day of daylight savings time (DST) change. Naive time in UTC.
        >>> date = dt.datetime(2020, 3, 29, 1, 15)
        >>> time_to_local_time(date, timezone_name='cet')
        Timestamp('2020-03-29 03:15:00')

    Example 3: Timezone has been set. Convert Egypt time to Central Europe time
        >>> date = dt.datetime(2020, 3, 2, 1, 15, tzinfo=zoneinfo.ZoneInfo('Africa/Cairo'))
        >>> time_to_local_time(date, timezone_name='CET')
        Timestamp('2020-03-02 00:15:00')

    Example 4: Pandas DatetimeIndex. Convert UTC to Sao Paulo, Brazil time
        >>> times = pd.date_range("2001-02-03 ", periods=3, freq='h') # Starts at 2001-02-03T00h00
        >>> time_to_local_time(times, timezone_name='America/Sao_Paulo')
        DatetimeIndex(['2001-02-02 22:00:00', '2001-02-02 23:00:00',
                       '2001-02-03 00:00:00'],
                      dtype='datetime64[ns]', freq=None)
    """
    if not isinstance(timezone_name, str):
        raise TypeError("`timezone_name` must be a `str`")

    if isinstance(times, dt.datetime):
        return _datetimeindex_to_local_time_tz_naive(pd.DatetimeIndex([times]), timezone_name)[0]
    else:
        return _datetimeindex_to_local_time_tz_naive(times, timezone_name)


def _ensure_datetimeindex(
    times: T.Union[dt.datetime, T.Iterable[dt.datetime], pd.DatetimeIndex] = None
) -> pd.DatetimeIndex:
    """
    Ensure a dt.datetime, iterable of dt.datetime, or pd.DateTimeIndex is returned as a pd.DateTimeIndex
    """
    if isinstance(times, pd.DatetimeIndex):
        return times
    elif isinstance(times, dt.datetime):
        return pd.DatetimeIndex([times], name="time")
    else:
        return pd.DatetimeIndex(times, name="time")


def _datetime_to_UTC(times: T.Union[dt.datetime, T.Iterable[dt.datetime], pd.DatetimeIndex] = None) -> pd.DatetimeIndex:
    """
    Ensure a datetime is timezone aware, set to UTC. `times` can be timezone-naive (where no `tz` has been set)
    or have a timezone set.

    Args:
        times: Datetimes - can be timezone naive or have timezone set to 'UTC'

    Returns:
        pd.DateTimeIndex that is timezone aware, set to UTC
    """
    times = _ensure_datetimeindex(times)

    if times.tzinfo is None:
        # Assume times are in UTC and localize time to UTC
        return times.tz_localize(tz="UTC")
    else:
        # Convert times to UTC
        return times.tz_convert("UTC")


def _datetimeindex_to_local_time_tz_aware(
    times: T.Union[dt.datetime, T.Iterable[dt.datetime], pd.DatetimeIndex] = None, timezone_name: str = None
) -> pd.DatetimeIndex:
    """
    Apply a timezone / daylight-savings-time (dst) offset to a datetime-like object. The Timestamps can be
    timezone-naive (no `tz` has been set), where the times are assumed to be in UTC, or have a timezone set.

    Args:
        times: Datetimes can be timezone naive or have timezone set.
        timezone_name: Name of the timezone, understood by `zoneinfo`

    Returns:
        Datetimes where the required offset has been applied to the time
    """
    # Ensure timezone is set to UTC
    times = _datetime_to_UTC(times)

    if timezone_name is None:
        # Times are assumed to be in UTC
        return times
    else:
        # Apply conversion from UTC to new timezone
        return times.tz_convert(timezone_name)


def _datetimeindex_to_local_time_tz_naive(
    times: T.Union[dt.datetime, T.Iterable[dt.datetime], pd.DatetimeIndex] = None, timezone_name: str = None
) -> pd.DatetimeIndex:
    """
    Apply a timezone / daylight-savings-time (dst) offset to a datetime-like object. The Timestamps are assumed
    to be in UTC - they can be timezone-naive (no `tz` has been set) or have a timezone set.

    Args:
        times: Datetimes where the time is in UTC - can be timezone naive or have timezone set to 'UTC'
        timezone_name: Name of the timezone, understood by `zoneinfo`

    Returns:
        Datetimes where the required offset has been applied to the time, and the datetimes are returned
        as timezone-naive.
    """
    local_times = _datetimeindex_to_local_time_tz_aware(times, timezone_name)

    # Make times timezone-naive
    local_times = local_times.tz_localize(None)

    return local_times


def utc_offset_in_hours(
    times: T.Union[dt.datetime, T.Iterable[dt.datetime], pd.DatetimeIndex, pd.Timestamp] = None,
    timezone_name: str = None,
    return_single_value: bool = True,
) -> T.Union[float, T.List[float]]:
    """
    Return the offset in (decimal) hours between UTC time and a local timezone, for a given datetime.
    Assumes all datetimes in `times` have the same timezone.

    Args:
        times: Timestamps, assumed to be in UTC (if timezone-naive or no timezone is set). Timestamps can
        be timezone-naive (no `tz` has been set), or have a timezone set.
        timezone_name: Name of the timezone, understood by `zoneinfo`

    Returns:
        Offset in decimal hours between the given timezone and UTC.
    """
    times = _datetimeindex_to_local_time_tz_aware(times, timezone_name)
    if len(times) == 1 or return_single_value:
        return times[0].utcoffset().total_seconds() / 3600
    else:
        return [time.utcoffset().total_seconds() / 3600 for time in times]


def _set_time_in_data(
    data: T.Union[xr.Dataset, xr.DataArray, pd.DataFrame, pd.Series] = None,
    time_dim: str = "time",
    new_times: pd.DatetimeIndex = None,
    set_time_to_midnight: bool = False,
    hours_to_subtract: float = None,
) -> T.Union[xr.Dataset, xr.DataArray, pd.DataFrame, pd.Series]:
    """
    Change the timestamps in the `time_dim` of `data`, optionally resetting the time to midnight (00:00:00),
    or subtracting some number of hours (`hours_to_subtract`), or replacing with a new pd.DatatimeIndex.

    Args:
        data: Input data.
        time_dim: Name of the time dimension/index/column in `data` that will be modified.
        new_times: Optional `pd.DatetimeIndex` that will replace the timestamps in `time_dim`. Takes precedence
        over all arguments.
        set_time_to_midnight: If `True`, reset the time part of the timestamps to midnight (00:00:00) on the same day.
        If `False`, do not modify the time part of the timestamps.
        hours_to_subtract: Optional number of hours that will be subtracted from the timestamps in `time_dim`. Takes
        precedence over `set_time_to_midnight`.

    Returns:
        The original `data` with modified timestamps in `time_dim`.
    """
    if new_times is None:  # If no replacement times are given

        if hours_to_subtract is not None:
            # Subtract some number of hours from the times, ignoring `set_time_to_midnight`
            new_times = get_time_from_data(data, time_dim=time_dim) - dt.timedelta(hours=hours_to_subtract)

        elif set_time_to_midnight is not True:
            # Nothing to do so return un-modified data
            return data

        else:
            # Reinitialize the time component of the datetime to midnight i.e. 00:00:00, ignoring `hours_to_subtract`
            new_times = get_time_from_data(data, time_dim=time_dim).normalize()

    return tp.utils.set_dim_values_in_data(data=data, values=new_times, dim=time_dim)


def data_to_local_time(
    data: T.Union[xr.Dataset, xr.DataArray, pd.DataFrame, pd.Series, np.ndarray, T.List, dt.datetime, pd.DatetimeIndex],
    timezone_name: str = None,
    time_dim: str = "time",
):
    """
    Convert and replace times in data to the correct local time, for a given country.
    By default returns original (unmodified) data.

    Args:
        data: Data with time coordinate/index/column. Datetimes are assumed to be in UTC if the timezone is not set.
        timezone_name: The name of the target timezone, understood by `zoneinfo`.
        Use `list_timezones()` to view all possible options.
        time_dim: The name of the time dimension/coordinate/column.

    Returns:
        Data with modified times

    Example 1: xarray.DataArray
        >>> da = xr.DataArray([1, 2], coords=[pd.date_range("2001-02-03 ", periods=2, freq='h')], dims="time")
        >>> da = data_to_local_time(da, 'America/Sao_Paulo') # Change to Sao Paulo, Brazil local time
        >>> da.indexes['time']
        DatetimeIndex(['2001-02-02 22:00:00', '2001-02-02 23:00:00'], dtype='datetime64[ns]', name='time', freq=None)

    Example 2: pandas.DataFrame
        >>> df = pd.DataFrame(
        ... {"time": pd.date_range("2019-03-15", freq="1D", periods=2), "val": [1, 2]}).set_index(["time"])
        >>> df = data_to_local_time(df, 'America/Sao_Paulo')
        >>> df.index.get_level_values('time')
        DatetimeIndex(['2019-03-14 21:00:00', '2019-03-15 21:00:00'], dtype='datetime64[ns]', name='time', freq=None)
    """
    if timezone_name is None:
        raise ValueError("`timezone_name` must be given")
    else:
        times = get_time_from_data(data, time_dim=time_dim)
        times = time_to_local_time(times, timezone_name)

        if isinstance(data, (np.ndarray, list, dt.datetime, pd.DatetimeIndex)):
            return times
        else:
            return tp.utils.set_dim_values_in_data(data=data, values=times, dim=time_dim)


def _resample_pandas_multiindex(
    data: T.Union[pd.DataFrame, pd.Series] = None,
    time_dim: str = "time",
    freq: str = "D",
    offset: T.Union[int, float] = None,
    closed: str = "left",
) -> T.Union[pd.DataFrame, pd.Series]:
    """
        Resample a pd.Dataframe/Series with a pd.Multiindex, where we do not reduce over other index levels,
        only `time_dim`.

        Args:
            data: Data to resample in time
            time_dim: Name of the time dimension/index/column in `data` that will be used to determine the grouping.
            freq: Resample frequency. Follows Pandas notation here
            https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#dateoffset-objects

    TODO        offset: Offset from base resampling
            closed:

        Returns:
            Groupby object with data grouped by the given frequency.
    """
    time_dim_level = data.index.names.index(time_dim)
    other_levels = np.delete(data.index.names, time_dim_level)
    grouped = data.groupby([pd.Grouper(level=time_dim, freq=freq, offset=offset, closed=closed), *other_levels])
    return grouped


def groupby_freq(
    data,
    time_dim: str = "time",
    freq: str = "D",
    closed: str = "left",
    day_start_hour: int = 0,
    other_grouping_columns: list = None,
):
    """
    Resample data in time.

    Args:
        data: Data to resample in time
        time_dim: Name of the time dimension/index/column in `data` that will be used to determine the grouping.
        freq: Resample frequency. Follows Pandas notation here
        https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#dateoffset-objects
        day_start_hour: The hour that is used as the start of the day (so 1 complete day is from
        `day_start_hour` to `day_start_hour` + 24h). Defaults to 0h.

    Returns:
        Groupby object with data grouped by the given frequency.

    Example: xr.Dataset / pd.DataFrame group by month
        >>> ds = xr.Dataset(data_vars={"var": (("lat", "lon", "time"), np.ones((1, 1, 100)))},
        ... coords={"lat": [1], "lon": [2], "time": pd.date_range("2022-01-01", periods=100)})
        >>> groupby_freq(ds, freq="M")
        DatasetResample, grouped over '__resample_dim__'
        4 groups with labels 2022-01-31, ..., 2022-04-30.
        >>> groupby_freq(ds['var'].to_dataframe(), freq="M").sum()  # doctest: +NORMALIZE_WHITESPACE
                             var
        time       lat lon
        2022-01-31 1   2    30.0
        2022-02-28 1   2    28.0
        2022-03-31 1   2    31.0
        2022-04-30 1   2    11.0
    """
    # The offset in hours to apply to the data to account for the day start hour
    offset = dt.timedelta(hours=day_start_hour)

    if isinstance(data, (xr.DataArray, xr.Dataset)):

        # Bug with xarray.Dataset.resample where it ignores the `base` argument, so a workaround is
        # implemented here, where the times are modified and `loffset` is used to re-label the time
        return _set_time_in_data(data, time_dim=time_dim, hours_to_subtract=day_start_hour).resample(
            {time_dim: freq}, loffset=str(day_start_hour) + "H", closed=closed
        )

    elif isinstance(data, (pd.Series, pd.DataFrame)):

        is_dim_in_index = tp.utils._dim_in_pandas_index(data.index, time_dim)

        if is_dim_in_index:
            is_multiindex = tp.utils._pandas_check_multiindex_type(data.index)

            if is_multiindex:
                return _resample_pandas_multiindex(
                    data=data, time_dim=time_dim, freq=freq, offset=offset, closed=closed
                )
            else:
                return data.groupby(pd.Grouper(level=time_dim, freq=freq, offset=offset, closed=closed))
        else:
            if isinstance(data, pd.DataFrame):
                # time_dim not in index, look in columns for DataFrames
                if time_dim in data.columns:

                    if other_grouping_columns is None:
                        return data.groupby(pd.Grouper(key=time_dim, freq=freq, offset=offset, closed=closed))
                    else:
                        other_grouping_columns = tp.utils.ensure_list(other_grouping_columns)
                        return data.groupby(
                            [pd.Grouper(key=time_dim, freq=freq, offset=offset, closed=closed), *other_grouping_columns]
                        )

            raise ValueError(f"time_dim=`{time_dim}` not found in data")
    else:
        data_types_str = ", ".join(
            str(i)
            for i in [
                xr.Dataset,
                xr.DataArray,
                pd.DataFrame,
                pd.Series,
            ]
        )
        raise TypeError(f"Data is of type {type(data)} but must be one of type: {data_types_str}")


def resample_time(
    data: T.Union[pd.DataFrame, pd.Series] = None,
    time_dim: str = "time",
    freq: str = "D",
    closed: str = "left",
    resample_method: str = "sum",
    day_start_hour: int = 0,
    other_grouping_columns: list = None,
):
    """
    Resample data in time.

    Args:
        data: Data to resample in time
        time_dim: Name of the time dimension/index/column in `data` that will be used to determine the aggregation.
        freq: Resample frequency. Follows Pandas notation here
        https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#dateoffset-objects
        resample_method: How to reduce/aggregate the data in each resampled group, one
        of 'mean', 'sum', 'max', 'min', 'cumsum'.
        day_start_hour: The hour that is used as the start of the day (so 1 complete day is from
        `day_start_hour` to `day_start_hour` + 24h). Defaults to 0h.

    Returns:
        Data resampled in time
    """
    grouped = groupby_freq(
        data,
        time_dim=time_dim,
        freq=freq,
        closed=closed,
        day_start_hour=day_start_hour,
        other_grouping_columns=other_grouping_columns,
    )

    # if resample_method == 'cumsum':

    #     time_index = tp.utils.get_dim_values_from_data(data, time_dim)
    #     # For xarray
    #     if isinstance(grouped, (xr.core.groupby.DatasetGroupBy, xr.core.groupby.DataArrayGroupBy)):
    #         cumsum = grouped.map(lambda x: x.cumsum(dim=dim, skipna=True, keep_attrs=True))
    #         return cumsum.reindex({time_dim: time_index})
    #     elif isinstance(grouped, (pd.core.groupby.generic.DataFrameGroupBy, pd.core.groupby.generic.SeriesGroupBy)):
    #         return grouped.cumsum(**kwargs)

    return tp.utils._call_resample_method(grouped, resample_method)


def rolling(
    data: T.Union[pd.Series, xr.DataArray] = None,
    n_periods: int = 3,
    min_periods: int = None,
    method: str = "sum",
    time_dim: str = "time",
) -> T.Union[pd.Series, xr.DataArray]:
    """
    Calculate a rolling `method` over a window with `n_periods`, with minimum of `min_periods` in the window.
    Works for both pd.DataFrame/Series and xr.DataArray/xr.Dataset

    Args:
        data: pd.Dataframe/Series or xr.Dataset/DataArray.
        n_periods: Number of periods to include in the sliding window, where the given `method` is applied to values
        from `time - n_periods` to `time`. i.e. `n_periods = 3` with `method='sum'` means the data at a given time will
        be summed with the previous 2 time steps

    """
    if min_periods is None:
        min_periods = n_periods

    if isinstance(data, pd.Series):
        rolling_data = data.rolling(n_periods, min_periods=n_periods)
    elif isinstance(data, xr.DataArray):
        rolling_data = data.rolling({time_dim: n_periods}, min_periods=n_periods)
    else:
        raise TypeError("data must be of type pd.Series or xr.DataArray")

    return tp.utils._call_resample_method(rolling_data, method, dim=time_dim)


################################################

# Dictionary to define the relative order of pandas time frequencies, where a lower value is higher frequency.
# FREQUENCY_DICT = {
#     "N": 0,
#     "U": 1,
#     "us": 1,
#     "L": 2,
#     "ms": 2,
#     "S": 3,
#     "T": 4,
#     "min": 4,
#     "H": 5,
#     "D": 6,
#     "W": 7,
#     "M": 8,
#     "Q": 9,
#     "A": 10,
#     "Y": 10,
# }


# def parse_date(date: T.Union[str, dt.datetime, pd.Timestamp] = None):
#     """
#     Parse a date string and return a dt.datetime.
#     By default return today's date at time 00:00
#     """
#     if date is None:
#         return dt.datetime.combine(dt.datetime.now(), dt.time.min)
#     elif isinstance(date, str):
#         return dateutil.parser.parse(date, fuzzy=True)
#     else:
#         return date


def yearly_date_range(
    start_time: dt.datetime = None,
    end_time: dt.datetime = None,
    reset_time: bool = True,
) -> T.List[dt.datetime]:
    """
    Generate a list of dates with a frequency of 1 year. If `reset_times==True`, the month, day
    and time are ignored (only using the year values). This is useful if you want to include the
    first and last year even if they are not complete.

    Args:
        start_time: Start date for range
        end_time: End date for range
        reset_time: Whether to ignore the month, day and time values of the dates

    Returns:
        List of dates

    Example 1: Generate a range of dates with yearly frequency, ignoring months, days and time

        >>> start_time = dt.datetime(1994, 11, 5, 7, 23)
        >>> end_time = dt.datetime(1997, 2, 7, 0)
        >>> yearly_date_range(start_time, end_time)  # doctest: +NORMALIZE_WHITESPACE
        [datetime.datetime(1994, 1, 1, 0, 0),
         datetime.datetime(1995, 1, 1, 0, 0),
         datetime.datetime(1996, 1, 1, 0, 0),
         datetime.datetime(1997, 1, 1, 0, 0)]

    Example 2: Generate a range of dates with yearly frequency, including months, days and time

        >>> start_time = dt.datetime(1994, 11, 5, 7, 23)
        >>> end_time = dt.datetime(1997, 2, 7, 0)
        >>> yearly_date_range(start_time, end_time, reset_time=False)  # doctest: +NORMALIZE_WHITESPACE
        [datetime.datetime(1994, 11, 5, 7, 23),
         datetime.datetime(1995, 11, 5, 7, 23),
         datetime.datetime(1996, 11, 5, 7, 23)]
    """
    if reset_time:
        # Only keep year values
        start_time = dt.datetime(start_time.year, 1, 1)
        end_time = dt.datetime(end_time.year, 1, 1)
    return list(rrule.rrule(freq=rrule.YEARLY, dtstart=start_time, until=end_time))


def select_time_range(data, start_time=None, end_time=None):
    """
    Select a time range in data for pd.DataFrame, pd.Series, xr.DataArray, xr.Dataset
    """
    if isinstance(data, (xr.DataArray, xr.Dataset)):
        return data.sel(time=slice(start_time, end_time))
    elif isinstance(data, pd.DataFrame):
        return data.loc[pd.IndexSlice[start_time:end_time, :], :]
    elif isinstance(data, pd.Series):
        return data.loc[pd.IndexSlice[start_time:end_time, :]]


# def freq_to_timedelta64(freq):
#     """
#     Transforms pd.resample method freq to timedelta64 in ns

#     Args:
#         freq (str): Resample freq (s, m, H, D, Y) or a multiple of them (for example "4D")
#     Returns:
#         (np.timedelta64) A timedelta64 in ns

#     example 1 : Day
#     >>> freq_to_timedelta64("D")
#     numpy.timedelta64(86400000000000,'ns')

#     example 2 : iHour
#     >>> freq_to_timedelta64("12H")
#     numpy.timedelta64(43200000000000,'ns')

#     example 3 : ihour
#     >>> freq_to_timedelta64("6h")
#     numpy.timedelta64(21600000000000,'ns')

#     example 4 : Month
#     >>> freq_to_timedelta64("M") is None
#     True
#     """
#     step, resample_freq = split_freq(freq)

#     if resample_freq is not None:
#         if resample_freq == "M":
#             logging.warning("M leads to ambiguous timedelta64. Unable to determine the timedelta64.")
#         elif len(resample_freq) == 1 and 0 < len(step):
#             if resample_freq == "H":
#                 resample_freq = resample_freq.lower()
#             return np.timedelta64(int(step), resample_freq).astype("timedelta64[ns]")
#         elif len(resample_freq) == 1 and len(step) == 0:
#             if resample_freq == "H":
#                 resample_freq = resample_freq.lower()
#             return np.timedelta64(1, resample_freq).astype("timedelta64[ns]")
#     else:
#         raise ValueError("Only s, m, h, H, D, Y are accepted as freq")


# def split_freq(freq):
#     """
#     Splits the resample freq into the step size and the time freq

#     example 1 : Day
#     >>> split_freq("D")
#     ('', 'D')

#     example 2 : iHour
#     >>> split_freq("6H")
#     ('6', 'H')

#     example 3 : iMonthStart
#     >>> split_freq("5MS")
#     ('5', 'M')

#     example 4 : Other
#     >>> split_freq("BS")
#     ('', None)
#     """
#     step = ""
#     resample_freq = None
#     for char in freq:
#         if char.isdigit():
#             step += char
#         elif char in "smhHDMY":
#             resample_freq = char
#             break
#     return step, resample_freq



from ._binning import digitize
from ._ranking import calculate_quantiles, rank
from ._stats import nearest_multiple, sigma_clip

__all__ = ["digitize", "calculate_quantiles", "rank", "nearest_multiple", "sigma_clip"]



import numpy as np

# Latitude
_MIN_LATITUDE_RADIANS = -1.5707963267948966
_MAX_LATITUDE_RADIANS = 1.5707963267948966

# Solar declination
_MIN_SOLAR_DECLINATION_RADIANS = -0.41015237421866746
_MAX_SOLAR_DECLINATION_RADIANS = 0.41015237421866746

# Sunset hour angle
_MIN_SUNSET_HOUR_ANGLE_RADIANS = 0.0
_MAX_SUNSET_HOUR_ANGLE_RADIANS = np.pi


def check_day_hours(hours):
    """
    Check that `hours` is in the range 0 to 24.
    """
    if hours is None:
        raise ValueError("`hours` must be provided")
    elif np.any(hours < 0) or np.any(hours > 24):
        raise ValueError("`hours` should be in range 0-24")


def check_day_of_year(day_of_year):
    """
    Check day of the year is valid, between 1-366.
    """
    if day_of_year is None:
        raise ValueError("`day_of_year` must be provided")
    elif np.any(day_of_year < 1) or np.any(day_of_year > 366):
        raise ValueError("Day of the year must be in range 1-366")


def check_latitude_radians(latitude):
    """
    Check latitude is within range -pi/2 to +pi/2 radians (-90 to +90 degrees)
    """
    if latitude is None:
        raise ValueError("`latitude` must be provided")
    elif np.any(latitude < _MIN_LATITUDE_RADIANS) or np.any(latitude > _MAX_LATITUDE_RADIANS):
        raise ValueError("latitude outside valid range {0} to {1}".format(_MIN_LATITUDE_RADIANS, _MAX_LATITUDE_RADIANS))


def check_solar_declination_radians(solar_declination):
    """
    Solar declination can vary between -23.5 and +23.5 degrees (in radians).

    See http://mypages.iit.edu/~maslanka/SolarGeo.pdf
    """
    if solar_declination is None:
        raise ValueError("`solar_declination` must be provided")
    elif np.any(solar_declination < _MIN_SOLAR_DECLINATION_RADIANS) or np.any(
        solar_declination > _MAX_SOLAR_DECLINATION_RADIANS
    ):
        raise ValueError(
            "solar declination outside valid range {0} to {1}".format(
                _MIN_SOLAR_DECLINATION_RADIANS, _MAX_SOLAR_DECLINATION_RADIANS
            )
        )


def check_sunset_hour_angle_radians(sunset_hour_angle):
    """
    Sunset hour angle has the range 0 to pi radians (0 to 180 degrees).

    See http://mypages.iit.edu/~maslanka/SolarGeo.pdf
    """
    if sunset_hour_angle is None:
        raise ValueError("`sunset_hour_angle` must be provided")
    elif np.any(sunset_hour_angle < _MIN_SUNSET_HOUR_ANGLE_RADIANS) or np.any(
        sunset_hour_angle > _MAX_SUNSET_HOUR_ANGLE_RADIANS
    ):
        raise ValueError(
            "`sunset_hour_angle` outside valid range {0} to {1}".format(
                _MIN_SUNSET_HOUR_ANGLE_RADIANS, _MAX_SUNSET_HOUR_ANGLE_RADIANS
            )
        )


def check_positive(values, string=None):
    """
    Check values are positive.

    Args:
        values: Array-like values to check.
        string: Optional string to pass to the raised value error, in the case of non-positive values.

    Returns:
        `True` if `values` is contains only positive numbers, otherwise `False`.
    """
    if values is None:
        raise ValueError("`values` must be provided")
    elif np.any(np.isnan(values)):
        raise ValueError("`values` contains `NaN`")
    elif np.any(values < 0.0):
        if string is not None:
            raise ValueError("{} has some non-positive values".format(string))
        else:
            raise ValueError("Non-positive values")



"""
Module containing functionality to estimate reference evapotransporation (ETo), sometimes referred to as
potential evapotranspiration (PET), for a grass reference crop using the FAO-56 Penman-Monteith equation.
"""
import typing as T

import numpy as np
import pandas as pd
import xarray as xr


def fao56_penman_monteith(
    data: T.Union[xr.Dataset, pd.DataFrame] = None,
    net_rad: T.Union[str, np.ndarray] = "net_rad",
    t: T.Union[str, np.ndarray] = "2t",
    ws: T.Union[str, np.ndarray] = "ws",
    svp: T.Union[str, np.ndarray] = "svp",
    avp: T.Union[str, np.ndarray] = "avp",
    delta_svp: T.Union[str, np.ndarray] = "delta_svp",
    psy: T.Union[str, np.ndarray] = "psy",
    shf: T.Union[str, np.ndarray] = None,
) -> T.Union[np.ndarray, pd.Series, xr.DataArray]:
    """
    Estimate reference evapotranspiration (ETo) from a hypothetical short grass reference surface using
    the FAO-56 Penman-Monteith equation. Based on equation 6 in Allen et al (1998).

    If `data==None`, the other arguments should all be `np.array` for each parameter. If `data` is
    a `pd.DataFrame` or `xr.Dataset`, the other arguments should be `str` of the names of each parameter
    in `data`.

    This is the method recommended by the Food and Agriculture Organisation of the United Nations (FAO)
    for estimating (ETo) for a short grass crop using limited meteorological data (see Allen et al, 1998).
    The FAO-56 Penman-Monteith equation requires site location, air temperature, humidity, radiation and
    wind speed data for daily, weekly, ten-day or monthly ETo calculations. It is important to verify the
    units of all input data.

    Args:
        data: pd.DataFrame or xr.Dataset containing columns/variables of all necessary parameters. If `None`,
        all other arguments must be of type `np.array` with values for each parameter.
        net_rad: Net radiation at crop surface [MJ m-2 day-1]. If necessary this can be estimated
        using ``solar_radiation.net_rad()``.
        t: Mean daily air temperature at 2 m height [deg Celcius].
        ws: Wind speed at 2 m height [m s-1]. If not measured at 2m, convert using ``wind_speed_at_2m()``.
        svp: Saturation vapour pressure [kPa]. Can be estimated using ``saturation_vapor_pressure_from_temperature()''.
        avp: Actual vapour pressure [kPa]. Can be estimated using a range of functions, in order of preference:
            1 - If dewpoint temperature data are available use ``actual_vapour_pressure_from_tdew()``.
            2 - If dry and wet bulb temperatures are available from a psychrometer
            use ``actual_vapour_pressure_from_twet_tdry()``.
            3 - If reliable minimum and maximum relative humidity data available
            use ``actual_vapour_pressure_from_rhmin_rhmax()``.
            4 - If measurement errors of relative humidity are large then use only maximum relative humidity
            using ``actual_vapour_pressure_from_rhmax()``.
            5 - If minimum and maximum relative humidity are not available but mean relative humidity is available
            then use ``actual_vapour_pressure_from_rhmean()`` (but this is less reliable than options 3 or 4).
            6 - If no data for the above are available then use ``actual_vapour_pressure_from_tmin()``.
            This function is less reliable in arid areas where it is recommended that 2 degrees Celsius is
            subtracted from the minimum temperature before it is passed to the function (following advice
            given in Annex 6 of Allen et al (1998).
        delta_svp: Slope of saturation vapour pressure curve [kPa degC-1]. Can be
        estimated using ``delta_saturation_vapor_pressure()``.
        psy: Psychrometric constant [kPa degC-1]. Can be estimatred using, in order of preference:
        ``psy_const_of_psychrometer()`` or ``psy_const()``.
        shf: Soil heat flux (G) [MJ m-2 day-1] (default is 0.0, which is reasonable for a daily or 10-day time
        steps). For monthly time steps ``shf`` can be estimated using ``monthly_soil_heat_flux()`` or
        ``monthly_soil_heat_flux2()``.

    Returns:
        Reference evapotranspiration (ETo) from a hypothetical grass reference surface [mm day-1].
        If `data` is `xr.Dataset`, returns a `xr.DataArray`. If `data` is `pd.DataFrame`, returns
        a `pd.Series`. Otherwise, return a `np.ndarray`.
    """
    # If using numpy arrays
    if data is None:
        return _fao56_penman_monteith(net_rad, t, ws, svp, avp, delta_svp, psy, shf)

    dim_order = ["lat", "lon", "time"]

    # Convert xr.Dataset to pd.Dataframe
    if isinstance(data, xr.Dataset):
        df = data.to_dataframe(dim_order=dim_order)
        df.index.names = dim_order
    elif isinstance(data, pd.DataFrame):
        if all([i in data.index.names for i in dim_order]):
            df = data.reorder_levels(dim_order, axis=0)
        else:
            df = data

    # Check all required parameters are given
    for name, argument_name in zip(
        ["net_rad", "t", "ws", "svp", "avp", "delta_svp", "psy"], [net_rad, t, ws, svp, avp, delta_svp, psy]
    ):
        if argument_name not in df.columns:
            raise ValueError(f"Variable name {name}={argument_name} not found in `data`")

    # initialize value of soil heat flux variable
    if shf is not None:
        if shf in df.columns:
            shf = df[shf]
        else:
            raise ValueError(f"Variable name `shf`={shf} not found in `data`")

    # eto is a pd.Series
    eto = _fao56_penman_monteith(df[net_rad], df[t], df[ws], df[svp], df[avp], df[delta_svp], df[psy], shf)
    eto.name = "eto"

    if isinstance(data, xr.Dataset):
        eto = eto.to_xarray()
        eto = eto.assign_attrs(
            {
                "units": "mm day-1",
                "description": "ETo reference evapotranspiration [mm day-1]",
            }
        )
        eto = eto.to_dataset()
    return eto


def _fao56_penman_monteith(net_rad, t, ws, svp, avp, delta_svp, psy, shf=0.0):
    """
    Estimate reference evapotranspiration (ETo) from a hypothetical short grass reference surface using
    the FAO-56 Penman-Monteith equation. Based on equation 6 in Allen et al (1998).

    This is the method recommended by the Food and Agriculture Organisation of the United Nations (FAO)
    for estimating (ETo) for a short grass crop using limited meteorological data (see Allen et al, 1998).
    The FAO-56 Penman-Monteith equation requires site location, air temperature, humidity, radiation and
    wind speed data for daily, weekly, ten-day or monthly ETo calculations. It is important to verify the
    units of all input data.

    Args:
        net_rad: Net radiation at crop surface [MJ m-2 day-1]. If necessary this can be estimated
        using ``solar_radiation.net_rad()``.
        t: Mean daily air temperature at 2 m height [deg Celcius].
        ws: Wind speed at 2 m height [m s-1]. If not measured at 2m, convert using ``wind_speed_at_2m()``.
        svp: Saturation vapour pressure [kPa]. Can be estimated using ``saturation_vapor_pressure_from_temperature()''.
        avp: Actual vapour pressure [kPa]. Can be estimated using a range of functions, in order of preference:
            1 - If dewpoint temperature data are available use ``actual_vapour_pressure_from_tdew()``.
            2 - If dry and wet bulb temperatures are available from a psychrometer
            use ``actual_vapour_pressure_from_twet_tdry()``.
            3 - If reliable minimum and maximum relative humidity data available
            use ``actual_vapour_pressure_from_rhmin_rhmax()``.
            4 - If measurement errors of relative humidity are large then use only maximum relative humidity
            using ``actual_vapour_pressure_from_rhmax()``.
            5 - If minimum and maximum relative humidity are not available but mean relative humidity is available
            then use ``actual_vapour_pressure_from_rhmean()`` (but this is less reliable than options 3 or 4).
            6 - If no data for the above are available then use ``actual_vapour_pressure_from_tmin()``.
            This function is less reliable in arid areas where it is recommended that 2 degrees Celsius is
            subtracted from the minimum temperature before it is passed to the function (following advice
            given in Annex 6 of Allen et al (1998).
        delta_svp: Slope of saturation vapour pressure curve [kPa degC-1]. Can be
        estimated using ``delta_saturation_vapor_pressure()``.
        psy: Psychrometric constant [kPa degC-1]. Can be estimatred using, in order of preference:
        ``psy_const_of_psychrometer()`` or ``psy_const()``.
        shf: Soil heat flux (G) [MJ m-2 day-1] (default is 0.0, which is reasonable for a daily or 10-day time
        steps). For monthly time steps ``shf`` can be estimated using ``monthly_soil_heat_flux()`` or
        ``monthly_soil_heat_flux2()``.

    Returns:
        Reference evapotranspiration (ETo) from a hypothetical grass reference surface [mm day-1].
    """
    if shf is None:
        shf = 0.0

    # Check all required parameters are given
    for name, variable in zip(
        ["net_rad", "t", "ws", "svp", "avp", "delta_svp", "psy", "shf"], [net_rad, t, ws, svp, avp, delta_svp, psy, shf]
    ):
        if variable is None:
            raise ValueError(f"Parameter {name} is not given")

    numerator = (0.408 * (net_rad - shf) * delta_svp) + ((psy * 891.3 * ws * (svp - avp)) / (t + 273))
    denominator = delta_svp + (psy * (1 + 0.3365 * ws))
    return numerator / denominator



import unittest

import numpy as np
import pandas as pd

import terrapyn as tp


class TestPandasToGeoPandas(unittest.TestCase):
    def test_geometry(self):
        df = pd.DataFrame({"lon": [1, 2], "lat": [3, 4], "value": [5, 6]})
        result = tp.utils.pandas_to_geopandas(df)
        self.assertEqual([(point.x, point.y) for point in result["geometry"].values], [(1.0, 3.0), (2.0, 4.0)])


class TestSetDimValuesInData(unittest.TestCase):

    df = pd.DataFrame(
        {
            "time": pd.DatetimeIndex(
                ["2019-03-15", "2019-03-16", "2019-03-17"], dtype="datetime64[ns]", name="time", freq=None
            ),
            "id": [123, 456, 789],
            "val": [1, 3, 5],
        }
    ).set_index("time")

    def test_replace_dataframe_index(self):
        expected = pd.DatetimeIndex(
            ["2011-01-01", "2011-01-02", "2011-01-03"], dtype="datetime64[ns]", name="time", freq=None
        )
        result = tp.utils.set_dim_values_in_data(self.df, expected, dim="time")
        pd.testing.assert_index_equal(result.index, expected)

    def test_replace_dataframe_multiindex(self):
        expected = pd.DatetimeIndex(
            ["2011-01-01", "2011-01-02", "2011-01-03"], dtype="datetime64[ns]", name="time", freq=None
        )
        result = tp.utils.set_dim_values_in_data(
            self.df.reset_index(drop=False).set_index(["time", "id"]), expected, dim="time"
        )
        pd.testing.assert_index_equal(result.index.get_level_values("time"), expected)

    def test_replace_dataframe_column(self):
        expected = [6, 7, 8]
        result = tp.utils.set_dim_values_in_data(self.df, expected, dim="val")
        self.assertEqual(result["val"].to_list(), expected)

    def test_replace_series_index(self):
        expected = pd.DatetimeIndex(
            ["2011-01-01", "2011-01-02", "2011-01-03"], dtype="datetime64[ns]", name="time", freq=None
        )
        result = tp.utils.set_dim_values_in_data(self.df["val"], expected, dim="time")
        pd.testing.assert_index_equal(result.index, expected)

    def test_replace_series_values(self):
        expected = [6, 7, 8]
        result = tp.utils.set_dim_values_in_data(self.df["val"], expected)
        self.assertEqual(result.to_list(), expected)

    def test_replace_dataset_dimension(self):
        expected = pd.DatetimeIndex(
            ["2011-01-01", "2011-01-02", "2011-01-03"], dtype="datetime64[ns]", name="time", freq=None
        )
        result = tp.utils.set_dim_values_in_data(self.df.to_xarray(), expected, dim="time")
        np.testing.assert_array_equal(result["time"].values, expected.values)

    def test_wrong_data_type(self):
        with self.assertRaises(TypeError):
            tp.utils.set_dim_values_in_data([1, 2, 3], [4, 5, 6])


class TestRemoveListElements(unittest.TestCase):

    input_list = ["item", 5, "foo", 3.14, True]

    def test_remove_element(self):
        result = tp.utils.remove_list_elements(self.input_list, ["item"])
        self.assertEqual(result, [5, "foo", 3.14, True])

    def test_missing_list(self):
        with self.assertRaises(ValueError):
            tp.utils.remove_list_elements(self.input_list, None)

    def test_string(self):
        result = tp.utils.remove_list_elements(self.input_list, "item")
        self.assertEqual(result, [5, "foo", 3.14, True])



import unittest

import xarray as xr

import terrapyn as tp


class TestChunkXarray(unittest.TestCase):

    ds = xr.open_dataset("tests/data/lat_10_lon_10_time_10_D_test_data.nc")

    def test_dataset_chunk_size_provided(self):
        result = tp.dask_utils.chunk_xarray(self.ds, "time", {"lat": 3, "lon": 3}).chunks
        expected = {"lat": (3, 3, 3, 1), "lon": (3, 3, 3, 1), "time": (10,)}
        self.assertEqual(result, expected)

    def test_dataset_all_auto_chunked(self):
        result = tp.dask_utils.chunk_xarray(self.ds).chunks
        expected = {"lat": (10,), "lon": (10,), "time": (10,)}
        self.assertEqual(result, expected)



from . import (
    conversion,
    dask_utils,
    indices,
    params,
    space,
    stats,
    time,
    utils,
    validation,
)

__all__ = ["time", "utils", "stats", "validation", "indices", "dask_utils", "conversion", "space", "params"]



import typing as T

import geopandas as gpd
import numpy as np
import pandas as pd
import shapely.geometry
import xarray as xr
from pyproj import Geod
from scipy.spatial import cKDTree


class BBox:
    """
    Object to define and store a bounding box. Coordinates can be user-defined by passing
    coordinates, or by passing a geometry. Bounding coordinates are stored as a shapely polygon,
    accessible with the `geom` attribute.

    By default, bounding box is set to the entire globe

    Attributes:
        min_lon (float): The minimum longitude
        max_lon (float): The maximum longitude
        min_lat (float): The minimum latitude
        max_lat (float): The maximum latitude
        geom (shapely.geometry): The shapely.geometry object that defines the bounding box.
        bounds: Returns the outer bounds of the geometry (a rectangle).
        WNES (tuple): Tuple of bounding coordinates in order of West, North, East, South.
        NWSE (tuple): Tuple of bounding coordinates in order of North, West, South, East.
    """

    def __init__(self, geometry=None, min_lon=-180, max_lon=180, min_lat=-90, max_lat=90):

        if geometry is None:
            self.geom = shapely.geometry.box(minx=min_lon, miny=min_lat, maxx=max_lon, maxy=max_lat)
        else:
            if not isinstance(geometry, shapely.geometry.base.BaseGeometry):
                raise TypeError("`geometry` must be a shapely geometry type")
            self.geom = geometry

        self.min_lon, self.min_lat, self.max_lon, self.max_lat = self.geom.bounds

    def __repr__(self):
        return (
            f"BBox(min_lon={self.min_lon}, max_lon={self.max_lon}, " f"min_lat={self.min_lat}, max_lat={self.max_lat})"
        )

    @property
    def bounds(self):
        return self.geom.bounds

    @property
    def WNES(self):
        return (self.min_lon, self.max_lat, self.max_lon, self.min_lat)

    @property
    def NWSE(self):
        return (self.max_lat, self.min_lon, self.min_lat, self.max_lon)

    @property
    def centroid(self):
        return self.geom.centroid

    @property
    def centroid_coords(self) -> T.Tuple[float, float]:
        """Coordinates of the centroid of bounding region, a tuple of (latitude, longitude)."""
        lon, lat = list(self.geom.centroid.coords)[0]
        return (lat, lon)

    def buffer(self, distance=0, resolution=16, cap_style=1, join_style=1, mitre_limit=5.0, single_sided=False):
        """
        Buffer/Expand BBox geometry following Shapely buffer method and return new BBox. Does not modify
        original BBox geometry. https://shapely.readthedocs.io/en/latest/manual.html#object.buffer
        """
        return BBox(
            geometry=self.geom.buffer(
                distance=distance,
                resolution=resolution,
                cap_style=cap_style,
                join_style=join_style,
                mitre_limit=mitre_limit,
                single_sided=single_sided,
            )
        )


def get_data_at_coords(
    data: T.Union[xr.Dataset, xr.DataArray] = None,
    lats: T.Iterable = None,
    lons: T.Iterable = None,
    point_names: T.Iterable = None,
    point_names_dim: str = "id",
    method: str = "nearest",
    lat_dim: str = "lat",
    lon_dim: str = "lon",
    time_dim: str = "time",
    ignore_nan: bool = False,
    as_dataframe: bool = True,
    **kwargs,
) -> T.Union[xr.Dataset, xr.DataArray, pd.DataFrame]:
    """
    Retrieve values for the specified coordinates from an xr.Dataset/Array, for all time steps.

    Args:
        data: Dataset/DataArray of source data.
        lat_dim: Name of latitude dimension in source data.
        lon_dim: Name of longitude dimension in source data.
        time_dim: Name of time dimension in source data.
        method: Method to use to select the values. If ``nearest`` and ``ignore_nan==True`` this uses a
        custom function, otherwise ``method`` is passed to `scipy.interpolate.interp1d`, where addition
        kwargs are passed to scipy via the `**kwargs` argument.
        point_names: Array of unique identifiers for points.
        point_names_dim: Name of the index for the `point_names`, default=`id`.
        lats: Array of latitudes. Not required if `metadata` is provided.
        lons: Array of longitudes. Not required if `metadata` is provided.
        as_dataframe: Whether to return data as a pandas.DataFrame or an xr.Dataset/Array

    Returns:
        Data at the given coordinates, either as a xr.Dataset/DataArray or pandas.DataFrame.

    Example: Retrieve the nearest non-NaN points

        >>> n_time = 5
        >>> data = np.full((n_time, 4, 4), np.nan)
        >>> data[:, 1, 1] = 1
        >>> data[:, 1, 2] = 2
        >>> data[:, 2, 1] = 3
        >>> data[:, 2, 2] = 4
        >>> data[2, 2, 2] = np.nan
        >>> da = xr.DataArray(data,
        ...                   coords={'lat':('lat', range(4)), 'lon': ('lon', range(4, 8)),
        ...                   'time': ('time', pd.date_range("01-01-2001", periods=n_time, freq='D'))},
        ...                   dims=['time', 'lat', 'lon'], name='test')
        >>> lats = [0, 3]
        >>> lons = [4, 6]
        >>> point_names = ['a', 'b']
        >>> get_data_at_coords(  # doctest: +NORMALIZE_WHITESPACE
        ...                    da, lats=lats, lons=lons, point_names=point_names,
        ...                    method='nearest', ignore_nan=True)
                           test
            time       id
            2001-01-01 a    1.0
                       b    4.0
            2001-01-02 a    1.0
                       b    4.0
            2001-01-03 a    1.0
                       b    3.0
            2001-01-04 a    1.0
                       b    4.0
            2001-01-05 a    1.0
                       b    4.0
    """
    if lats is None or lons is None:
        raise ValueError("Must provide `lats` and `lons`")

    # Retrieve nearest non-NaN values, for each coordinate that is not lat or lon (so time etc.)
    if method == "nearest" and ignore_nan:

        # Convert dataarray to dataframe
        points = data.to_dataframe().dropna()

        non_lat_lon_index_names = list(points.index.names.difference([lat_dim, lon_dim]))

        points = points.groupby(non_lat_lon_index_names).apply(
            get_nearest_point,
            lats=lats,
            lons=lons,
            lat_dim=lat_dim,
            lon_dim=lon_dim,
            return_distance=False,
            reset_index=True,
            index_name=point_names_dim,
        )

    # Retrieve nearest values (including NaN) for each coordinate that is not lat or lon (so time etc.)
    else:
        lat_idx = xr.DataArray(lats, dims=point_names_dim)
        lon_idx = xr.DataArray(lons, dims=point_names_dim)
        points = data.interp({lat_dim: lat_idx, lon_dim: lon_idx}, method=method, kwargs=kwargs)

        # Convert xr.DataArray to pd.DataFrame
        points = points.to_dataframe()

    # Drop lat and lon columns
    if lat_dim in points.columns:
        points.drop(columns=lat_dim, inplace=True)
    if lon_dim in points.columns:
        points.drop(columns=lon_dim, inplace=True)

    # Rename labels for point coordinates to `point_names`
    if point_names is not None:
        if isinstance(points.index, pd.MultiIndex):
            points.index = points.index.set_levels(point_names, level=point_names_dim)
        else:
            points.index = pd.Index(point_names, name=point_names_dim)

    if as_dataframe:
        # Return dataframe with `time_dim` and `point_names_dim` as the index (if both exist),
        # with all other coordinates reset to columns.
        index_coords_to_reset = list(points.index.names.difference([time_dim, point_names_dim]))
        if len(index_coords_to_reset) > 0:
            points = points.reset_index(index_coords_to_reset)
        if time_dim in points.index.names:
            dim_order = [time_dim, point_names_dim]
            points.index = points.index.reorder_levels(dim_order)
        else:
            dim_order = [point_names_dim]
        points = points.sort_index(level=dim_order)
    else:
        points = points.to_xarray()
    return points


def _geodesic_distances_between_point_and_endpoints(point: T.Tuple[float, float], endpoints):
    """
    Return the geodesic distances in kilometres (km) between `point` and the points in `endpoints`.
    `point` and `endpoints` have column order of (lat, lon)
    """
    endpoints = np.asarray(endpoints)

    # Check array shapes are valid
    if len(point) != 2:
        raise ValueError("`point` must be a tuple of (lat, lon)")
    if len(endpoints.shape) != 2 & endpoints.shape[1] != 2:
        raise ValueError("`endpoints` must be an n x 2 array")

    point = np.full(endpoints.shape, point)
    geod = Geod(ellps="WGS84")
    distance_km = geod.inv(point[:, 1], point[:, 0], endpoints[:, 1], endpoints[:, 0])[2] / 1000
    return distance_km


def _index_and_distance_of_nearest_point(point, endpoints):
    """
    Return the index and geodesic distance in km of the nearest point in `endpoints`, to the coordinate `point`.
    `point` and `endpoints` have column order of (lat, lon)
    """
    distances_km = _geodesic_distances_between_point_and_endpoints(point, endpoints)
    index_of_nearest = np.argmin(distances_km)
    return index_of_nearest, distances_km[index_of_nearest]


def get_nearest_point(
    df: pd.DataFrame,
    lats: T.Union[float, T.Iterable],
    lons: T.Union[float, T.Iterable],
    lat_dim: str = "lat",
    lon_dim: str = "lon",
    method: str = "geodesic",
    return_distance: bool = True,
    reset_index: bool = True,
    index_name: str = "point",
) -> pd.DataFrame:
    """
    Return nearest points from a pandas.DataFrame

    Args:
        df: Dataframe with values at coordinates
        lats: Array of latitudes at which to select nearest points.
        lons: Array of longitudes at which to select nearest points.
        lat_dim: Name of latitude index/column in data.
        lon_dim: Name of longitude index/colimn in data.
        method: Method to use to retrieve points, one of 'kdtree' or 'geodesic'. If 'geodesic', the
        distance can optionally be returned with `return_distance==True`
        return_distance: If `True`, add a column of distance (in km) from the source point to the target.
        Only applies when `method=='geodesic'`.
        reset_index: Whether to reset the dataframe index values or not.
        index_name: The name for the output index.

    Returns:
        A dataframe with the nearest points to the desired `lats` and `lons` coordinates
    """
    if isinstance(lats, (float, int)):
        lats = [lats]
    if isinstance(lons, (float, int)):
        lons = [lons]

    # order of (lat, lon)
    points = np.column_stack([lats, lons])
    endpoints = np.column_stack(get_lat_lon_from_data(df, lat_name=lat_dim, lon_name=lon_dim, unique=False))

    if method == "kdtree":
        tree = cKDTree(endpoints)
        _, indices = tree.query(points, k=1)

    elif method == "geodesic":
        # For each point, find the nearest endpoint and add the index to a list
        indices = []
        distances = []
        for p in points:
            index_of_nearest, distance_of_nearest = _index_and_distance_of_nearest_point(p, endpoints)
            indices.append(index_of_nearest)
            distances.append(distance_of_nearest)

    df = df.iloc[indices].copy()

    if return_distance and method == "geodesic":
        df["distance_km"] = distances

    if reset_index:
        df = df.reset_index(drop=True)
    if index_name is not None:
        df.index = df.index.rename(index_name)

    return df


def get_lat_lon_from_data(
    data: T.Union[xr.Dataset, xr.DataArray, pd.DataFrame, pd.Series, gpd.GeoDataFrame] = None,
    lon_name: str = "lon",
    lat_name: str = "lat",
    unique: bool = True,
) -> T.Tuple[np.ndarray, np.ndarray]:
    """
    Extract latitude and longitude coordinates from data

    Args:
        data: Input data
        lon_name: Name of longitude dimension/column
        lat_name: Name of latitude dimension/column
        unique: Return only unique pairs of lat and lon - useful for pandas.DataFrame where coordinates are repeated

    Returns:
        Tuple of arrays of lat and lon
    """
    if isinstance(data, (xr.Dataset, xr.DataArray)):
        lats = data[lat_name].values
        lons = data[lon_name].values
    elif isinstance(data, (pd.DataFrame, pd.Series, gpd.GeoDataFrame)):
        if lon_name in data.index.names and lat_name in data.index.names:
            if unique:
                lats = data.index.unique(lat_name).values
                lons = data.index.unique(lon_name).values
            else:
                lats = data.index.get_level_values(lat_name).values
                lons = data.index.get_level_values(lon_name).values
        elif lon_name in data.columns and lat_name in data.columns:
            if unique:
                lats = data[lat_name].unique()
                lons = data[lon_name].unique()
            else:
                lats = data[lat_name].values
                lons = data[lon_name].values
        else:
            raise ValueError("{} and {} were not found in the data".format(lat_name, lon_name))
    return lats, lons


def crop_to_bbox(
    data: T.Union[xr.Dataset, xr.DataArray, pd.DataFrame, pd.Series, gpd.GeoDataFrame] = None,
    bbox: BBox = None,
    lon_name: str = "lon",
    lat_name: str = "lat",
    geometry_name: str = "geometry",
) -> T.Union[xr.Dataset, xr.DataArray]:
    """
    Return all data within the given bounding box. Accepts xarray dataset/array or pandas/geopandas dataframe.
    Latitude can be ordered positive to negative, or negative to positive.

    Args:
        data: Input data
        BBox: Bounding box for data. If `None`, data is returned unaltered
        lon_name: Name of longitude dimension/column
        lat_name: Name of latitude dimension/column
        geometry_name: Name of column containing coordinates of shapely point objects in geopandas.GeoDataFrame

    Returns:
        Subset of data that is within the given bounding box
    """
    if bbox is None:
        return data
    else:
        # Add additional padding to bbox to correct for rounding errors, where units are assumed
        # to be in degrees, where distance is ~1cm on Earth's surface
        bbox = bbox.buffer(distance=0.0000001)

    if isinstance(data, (xr.Dataset, xr.DataArray)):
        # Select longitude range
        data = data.sel({lon_name: slice(bbox.min_lon, bbox.max_lon)})

        if data[lat_name][0] < data[lat_name][-1]:
            # Latitude Ordered Negative to Positive
            return data.sel({lat_name: slice(bbox.min_lat, bbox.max_lat)})
        else:
            # Latitude Ordered Positive to Negative
            return data.sel({lat_name: slice(bbox.max_lat, bbox.min_lat)})

    elif isinstance(data, (pd.DataFrame, pd.Series, gpd.GeoDataFrame)):

        # If data is geopandas.geodataframe, first try to select using shapely geometry
        if isinstance(data, gpd.GeoDataFrame):
            if geometry_name in data.columns:
                indexes_within_bbox = data[geometry_name].within(bbox.geom)
                return data.loc[indexes_within_bbox]

        # If data is pandas dataframe/series or geopandas dataframe without shapely geometry
        lats, lons = get_lat_lon_from_data(data, lon_name=lon_name, lat_name=lat_name, unique=False)

        # Generate shapely points for each coordinate and find points within bbox
        coordinates = gpd.points_from_xy(lons, lats)
        indexes_within_bbox = coordinates.within(bbox.geom)
        return data.loc[indexes_within_bbox]
    else:
        raise ValueError("Coordinates not found. Either `lon_name` and `lat_name` or `geometry_name` must be provided")


def match_data_bbox(
    reference_data: T.Union[xr.Dataset, xr.DataArray, pd.DataFrame, pd.Series, gpd.GeoDataFrame] = None,
    input_data: T.Union[xr.Dataset, xr.DataArray, pd.DataFrame, pd.Series, gpd.GeoDataFrame] = None,
    ref_lat_name: str = "lat",
    ref_lon_name: str = "lon",
    input_lat_name: str = "lat",
    input_lon_name: str = "lon",
    geometry_name: str = "geometry",
) -> T.Union[xr.Dataset, xr.DataArray, pd.DataFrame, pd.Series, gpd.GeoDataFrame]:
    """
    Crop the `input_data` to match the bounding box of the `output_data`. The data type of `input_data` is not modified.

    Args:
        reference_data: Reference data with coordinates that will be used to crop `input_data`.
        input_data: Input data that will be cropped.
        ref_lat_name: Name of latitude coordinate in `reference_data`.
        ref_lon_name: Name of longitude coordinate in `reference_data`.
        input_lat_name: Name of latitude coordinate in `input_data`.
        input_lon_name: Name of longitude coordinate in `input_data`.
        geometry_name: Name of column containing coordinates of shapely point objects if the data
        type is a geopandas.GeoDataFrame.

    Returns:
        Cropped input data.
    """
    output_bbox = bbox_from_data(reference_data, lat_name=ref_lat_name, lon_name=ref_lon_name)

    # Add additional padding to bbox to correct for rounding errors, ~1cm on Earth's surface
    output_bbox = output_bbox.buffer(distance=0.0000001)

    return crop_to_bbox(input_data, output_bbox, lon_name=input_lon_name, lat_name=input_lat_name)


def bbox_from_data(
    data: T.Union[xr.Dataset, xr.DataArray, pd.DataFrame, pd.Series, gpd.GeoDataFrame] = None,
    lon_name: str = "lon",
    lat_name: str = "lat",
    use_pixel_bounds: bool = False,
    decimals: int = 9,
) -> BBox:
    """
    Generate BBox from input data.

    Args:
        data: Input data
        lon_name: Name of longitude dimension/column
        lat_name: Name of latitude dimension/column
        use_pixel_bounds: If `True` the boundary will be the bounds of the outermost coordinates in the
        data, otherwise the bounds will be the pixel centers.
        decimals: Number of decimal places to return (using `np.around`). A value of `9` is
        around 1 cm on the Earth's surface if the coordinates are in degrees.

    Returns:
        BBox Subset of data that is within the given bounding box
    """
    lats, lons = get_lat_lon_from_data(data, lon_name, lat_name, unique=True)

    if use_pixel_bounds:
        lats = get_coordinate_bounds_from_centers(lats)
        lons = get_coordinate_bounds_from_centers(lons)

    min_lat = np.around(lats.min(), decimals=decimals)
    max_lat = np.around(lats.max(), decimals=decimals)
    min_lon = np.around(lons.min(), decimals=decimals)
    max_lon = np.around(lons.max(), decimals=decimals)

    return BBox(max_lat=max_lat, min_lon=min_lon, min_lat=min_lat, max_lon=max_lon)


def generate_grid(
    bbox: BBox = None,
    resolution: float = 1.0,
    min_lat: float = -90,
    max_lat: float = 90,
    min_lon: float = -180,
    max_lon: float = 180,
    variable_name: str = "var",
    fill_value: T.Union[int, float] = None,
    return_dataset: bool = False,
) -> T.Union[xr.DataArray, xr.Dataset]:
    """
    Makes an xr.DataArray/xr.Dataset grid with optional variable and fill value.

    Args:
        resolution: Resolution (in degrees).
        bbox: Optional Bounding box for data. Takes precedence over other coordinates.
        min_lat: Lower latitude.
        max_lat: Upper latitude.
        min_lon: Lower longitude.
        max_lon: Upper longitude.
        variable_name: Optional variable name.
        fill_value: Value used to fill the grid. Defaults to `np.nan` if `fill_value` is `None`.
        return_dataset: If True return an xr.Dataset, otherwise return an xr.DataArray.

    Returns:
        DataArray/set with the given coordinates, with optional fill value and variable name.
    """
    if bbox:
        max_lat, min_lon, min_lat, max_lon = bbox.NWSE
    lats = np.arange(min_lat, max_lat + resolution, resolution)
    lons = np.arange(min_lon, max_lon + resolution, resolution)

    if fill_value:
        # Generate data
        data = np.full(shape=(lats.shape[0], lons.shape[0]), fill_value=fill_value)
    else:
        # Return an empty xr.DataArray
        data = np.nan
    da = xr.DataArray(data=data, coords={"lat": lats, "lon": lons}, dims=["lat", "lon"], name=variable_name)

    if return_dataset:
        return da.to_dataset(name=variable_name)
    else:
        return da


def points_within_radius(
    df: pd.DataFrame = None,
    point: T.Tuple[float, float] = None,
    radius_km: float = 100,
    lat_col: str = "lat",
    lon_col: str = "lon",
    return_distance: bool = True,
) -> pd.DataFrame:
    """
    Return a subset of dataframe for coordinates that lie within the geodesic distance
    between coordinates (for accurate distances on Earth's surface).

    Args:
        df: Dataframe with unique identifier and coordinates
        point: Coordinate of source point in order (lat, lon)
        lat_col: Name of latitude column in dataframe.
        lon_col: Name of longitude column in dataframe.
        radius_km: Radius of circle in kilometres (km).
        return_distance: If `True`, add a column of distance (in km) from the source point to the target.

    Returns:
        Subset of dataframe with points inside the given radius.
    """
    endpoints = np.column_stack(get_lat_lon_from_data(df, lat_name=lat_col, lon_name=lon_col, unique=False))
    distances_km = _geodesic_distances_between_point_and_endpoints(point, endpoints)
    indices_within_radius = distances_km < radius_km
    df_within_radius = df.iloc[indices_within_radius].copy()
    if return_distance:
        df_within_radius["distance_km"] = distances_km[indices_within_radius]
    return df_within_radius


def add_coordinate_bounds(
    data: T.Union[xr.Dataset, xr.DataArray], lat_name: str = "lat", lon_name: str = "lon"
) -> T.Union[xr.Dataset, xr.DataArray]:
    """
    Add coordinate bounds 'lon_b' and 'lat_b' to an input dataset/array

    Args:
        data: Dataset/Array containing coordinates `lat_name` and `lon_name`
        lat_name: Name of the lat coordinate
        lon_name: Name of the lon coordinate

    Returns:
        The original data with new coordinates of `lat_b` and `lon_b` that give
        the boundaries of the input coordinates.
    """
    lons = data[lon_name].values
    lats = data[lat_name].values
    lat_bounds = get_coordinate_bounds_from_centers(lats)
    lon_bounds = get_coordinate_bounds_from_centers(lons)
    data["lon_b"] = lon_bounds
    data["lat_b"] = lat_bounds
    return data


def get_coordinate_bounds_from_centers(coords: np.ndarray = None) -> np.ndarray:
    """
    Return the boundaries (edges) of each cell/pixel from an array of center values, where we
    assume the difference between the first 2 values is the step size for all coordinates.
    This is useful to generate `lon_b` and `lat_b` coordinates for regridding purposes.

    Args:
        coords: Array of coordinates, the centers of the pixels/cells.

    Returns:
        Array of the coordinates boundaries, where there are `len(coords) + 1` elements.
    """
    step = coords[1] - coords[0]
    pad = step / 2.0
    bounds = np.linspace(coords[0] - pad, coords[-1] + pad, len(coords) + 1)
    return bounds


def get_coordinate_centers_from_bounds(coord_bounds: np.ndarray = None) -> np.ndarray:
    """
    Return the center of the pixels/cells from an array of pixel/cell bounds.

    Args:
        coord_bounds: Array of the boundaries of pixels/cells

    Returns:
        Array of the coordinate centers, where there are `len(coord_bounds) - 1` elements.
    """
    step = coord_bounds[1] - coord_bounds[0]
    pad = step / 2.0
    centers = coord_bounds[:-1] + pad
    return centers


def group_points_by_grid(
    df: pd.DataFrame = None,
    lat_col: str = "lat",
    lon_col: str = "lon",
    id_col: str = "id",
    cellsize: float = 5,
    return_cell_bounds: bool = True,
    lat_bounds: T.Iterable = None,
    lon_bounds: T.Iterable = None,
) -> pd.DataFrame:
    """
    Group points together based on location, using a grid. Each square in the grid has length
    'cellsize' in degrees, and the grid's bounds are defined by the outermost point coordinates.

    Args:
        df: Dataframe with point names and coordinates.
        lat_col: Name of latitude column in dataframe.
        lon_col: Name of longitude column in dataframe.
        id_col: Name of id (station name etc.) column in dataframe.
        cellsize: Size of each grid square in degrees, in which to group the points.
        return_cell_bounds: If `True` return additionally return arrays of the grid cell bounds
        lat_bounds: Array of grid boundaries for latitudes. If given, takes precedence over `cellsize`.
        lon_bounds: Array of grid boundaries for longitudes. If given, takes precedence over `cellsize`.

    Returns:
        pd.Series of groups of points (with integer labels) and a list of points in that group
    """
    lats, lons = get_lat_lon_from_data(df, lat_name=lat_col, lon_name=lon_col, unique=True)

    if lat_bounds is None and lon_bounds is None:
        min_lat, max_lat = lats.min(), lats.max()
        min_lon, max_lon = lons.min(), lons.max()

        # grid cells should extend cellsize * 0.5 beyond the outermost stations so stations are centered in cells
        pad = cellsize / 2.0

        lat_bins = np.arange(min_lat - pad, max_lat + pad + cellsize, cellsize)
        lon_bins = np.arange(min_lon - pad, max_lon + pad + cellsize, cellsize)
    else:
        lat_bins = lat_bounds
        lon_bins = lon_bounds

    lat_labels = np.digitize(lats, lat_bins)
    lon_labels = np.digitize(lons, lon_bins)

    groups = df.groupby([lat_labels, lon_labels])[id_col].apply(list)
    groups = groups.reset_index(drop=True).rename("group")

    if return_cell_bounds:
        return groups, lat_bins, lon_bins
    else:
        return groups
